/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.antragraum;

import java.time.ZonedDateTime;
import java.util.List;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.core.task.TaskExecutor;

import com.google.protobuf.ByteString;

import de.ozgcloud.apilib.file.OzgCloudFile;
import de.ozgcloud.common.binaryfile.GrpcBinaryFileServerDownloader;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.NachrichtenManagerConfiguration;
import de.ozgcloud.nachrichten.common.vorgang.VorgangService;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.RequiredArgsConstructor;
import net.devh.boot.grpc.server.service.GrpcService;

@GrpcService
@RequiredArgsConstructor
@ConditionalOnProperty(AntragraumProperties.PROPERTY_ANTRAGSRAUM_URL)
class AntragraumGrpcService extends AntragraumServiceGrpc.AntragraumServiceImplBase {

	static final int CHUNK_SIZE = 256 * 1024;

	private final AntragraumService service;
	private final AntragraumNachrichtMapper mapper;
	private final RueckfrageMapper rueckfrageMapper;
	private final AttachmentFileRequestMapper attachmentFileRequestMapper;
	@Qualifier(NachrichtenManagerConfiguration.NACHRICHTEN_OZG_CLOUD_FILE_MAPPER)
	private final OzgCloudFileMapper ozgCloudFileMapper;
	private final TaskExecutor taskExecutor;

	private final VorgangService vorgangService;

	@Override
	public void findRueckfragen(GrpcFindRueckfragenRequest request, StreamObserver<GrpcFindRueckfragenResponse> streamObserver) {
		var rueckfragen = service.findRueckfragen(request.getSamlToken()).map(mapper::toGrpc);

		streamObserver.onNext(buildFindRueckfragenResponse(rueckfragen));
		streamObserver.onCompleted();
	}

	private GrpcFindRueckfragenResponse buildFindRueckfragenResponse(Stream<GrpcRueckfrageHead> rueckfragen) {
		return GrpcFindRueckfragenResponse.newBuilder().addAllRueckfrageHead(rueckfragen.toList()).build();
	}

	@Override
	public void sendRueckfrageAnswer(GrpcSendRueckfrageAnswerRequest request, StreamObserver<GrpcSendRueckfrageAnswerResponse> streamObserver) {
		var answer = mapper.fromRueckfrageAnswer(request.getAnswer());

		answer = enrichRueckfrageAnswer(answer, getRueckfrageByNachricht(request.getSamlToken(), request.getAnswer().getRueckfrageId()));
		var commandId = service.sendRueckfrageAnswer(request.getSamlToken(), request.getAnswer().getRueckfrageId(), answer);

		streamObserver.onNext(GrpcSendRueckfrageAnswerResponse.newBuilder().setCommandId(commandId).build());
		streamObserver.onCompleted();
	}

	RueckfrageHead getRueckfrageByNachricht(String samlToken, String postfachNachrichtId) {
		return service.findRueckfragen(samlToken)
				.filter(rueckfrage -> StringUtils.equals(rueckfrage.getId(), postfachNachrichtId))
				.findFirst()
				.orElseThrow(() -> new TechnicalException(String.format("No rueckfrage found for answer with id %s", postfachNachrichtId)));
	}

	PostfachNachricht enrichRueckfrageAnswer(PostfachNachricht answer, RueckfrageHead rueckfrage) {
		return answer.toBuilder()
				.vorgangId(rueckfrage.getVorgangId())
				.postfachAddress(rueckfrage.getPostfachAddress())
				.sentAt(ZonedDateTime.now())
				.build();
	}

	@Override
	public void getRueckfrage(GrpcGetRueckfrageRequest request, StreamObserver<GrpcGetRueckfrageResponse> responseObserver) {
		var samlToken = request.getSamlToken();
		var rueckfrage = service.getRueckfrage(samlToken, request.getId());

		var enhancedRueckfrage = enhanceRueckfrage(rueckfrage, samlToken);

		responseObserver.onNext(buildGetRueckfrageResponse(enhancedRueckfrage));
		responseObserver.onCompleted();
	}

	GrpcRueckfrage enhanceRueckfrage(PostfachNachricht rueckfrage, String samlToken) {
		var vorgang = vorgangService.getVorgang(rueckfrage.getVorgangId());

		var answers = service.findAnswers(samlToken, rueckfrage.getId()).toList();
		var firstAnswerSentAt = service.getFirstSentAt(answers.stream());

		var rueckfrageWithAnswer = addAnswers(rueckfrageMapper.toRueckfrage(firstAnswerSentAt, vorgang, rueckfrage), answers);
		return rueckfrageWithAnswer.toBuilder().setAccessible(service.isAccessible(samlToken, vorgang.getTrustLevel())).build();
	}

	GrpcRueckfrage addAnswers(GrpcRueckfrage rueckfrage, List<PostfachNachricht> answers) {
		return rueckfrage.toBuilder().addAllAnswers(answers.stream().map(mapper::toRueckfrageAnswer).toList()).build();
	}

	GrpcGetRueckfrageResponse buildGetRueckfrageResponse(GrpcRueckfrage rueckfrage) {
		return GrpcGetRueckfrageResponse.newBuilder().setRueckfrage(rueckfrage).build();
	}

	@Override
	public void getAttachmentContent(GrpcGetAttachmentContentRequest request, StreamObserver<GrpcGetAttachmentContentResponse> responseObserver) {
		buildAttachmentDownloader(attachmentFileRequestMapper.fromContentRequest(request), responseObserver).start();
	}

	GrpcBinaryFileServerDownloader<GrpcGetAttachmentContentResponse> buildAttachmentDownloader(AttachmentFileRequest request,
			StreamObserver<GrpcGetAttachmentContentResponse> responseObserver) {
		return GrpcBinaryFileServerDownloader.<GrpcGetAttachmentContentResponse>builder()
				.callObserver((CallStreamObserver<GrpcGetAttachmentContentResponse>) responseObserver)
				.taskExecutor(taskExecutor)
				.downloadConsumer(outputStream -> service.getAttachmentContent(request, outputStream))
				.chunkBuilder(this::buildAttachmentChunkResponse)
				.build();
	}

	GrpcGetAttachmentContentResponse buildAttachmentChunkResponse(ByteString chunk) {
		return GrpcGetAttachmentContentResponse.newBuilder().setFileContent(chunk).build();
	}

	@Override
	public void getAttachmentMetadata(GrpcGetAttachmentMetadataRequest request, StreamObserver<GrpcGetAttachmentMetadataResponse> responseObserver) {
		var attachment = service.getAttachmentMetadata(attachmentFileRequestMapper.fromMetadataRequest(request));

		responseObserver.onNext(buildAttachmentMetadataResponse(attachment));

		responseObserver.onCompleted();
	}

	private GrpcGetAttachmentMetadataResponse buildAttachmentMetadataResponse(OzgCloudFile attachment) {
		return GrpcGetAttachmentMetadataResponse.newBuilder()
				.setFileMetadata(ozgCloudFileMapper.toMetadata(attachment))
				.build();
	}
}
