/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.antragraum;

import static java.util.Objects.*;

import java.io.OutputStream;
import java.time.ZonedDateTime;
import java.util.Comparator;
import java.util.Optional;
import java.util.stream.Stream;

import jakarta.annotation.PostConstruct;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.opensaml.saml.saml2.core.Response;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.stereotype.Service;

import de.ozgcloud.apilib.common.datatypes.GenericId;
import de.ozgcloud.apilib.common.errorhandling.NotFoundException;
import de.ozgcloud.apilib.file.OzgCloudFile;
import de.ozgcloud.apilib.file.OzgCloudFileId;
import de.ozgcloud.apilib.file.OzgCloudFileService;
import de.ozgcloud.nachrichten.NachrichtenManagerConfiguration;
import de.ozgcloud.nachrichten.NachrichtenManagerProperties;
import de.ozgcloud.nachrichten.common.vorgang.Vorgang;
import de.ozgcloud.nachrichten.common.vorgang.VorgangService;
import de.ozgcloud.nachrichten.postfach.AttachedItemService;
import de.ozgcloud.nachrichten.postfach.PersistPostfachNachrichtService;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtMapper;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Service
@RequiredArgsConstructor
@ConditionalOnProperty(AntragraumProperties.PROPERTY_ANTRAGSRAUM_URL)
public class AntragraumService {

	private static final String POSTFACH_NACHRICHT = "PostfachNachricht";

	static final String BAYERN_ID_SERVICE_KONTO_TYPE = "BAYERN_ID";

	static final String USER_NOTIFICATION_TEMPLATE = """
			Guten Tag,

			bei der Bearbeitung Ihres Antrags hat sich eine Rückfrage ergeben. Bitte beantworten Sie diese im Antragsraum:

			%s

			Vielen Dank,
			Ihre digitale Verwaltung
			""";

	private final PersistPostfachNachrichtService postfachNachrichtService;
	private final Saml2Verifier verifier;
	private final Saml2Parser parser;
	private final Saml2Decrypter decrypter;

	private final AntragraumProperties properties;
	private final NachrichtenManagerProperties nachrichtenManagerProperties;

	@Qualifier(NachrichtenManagerConfiguration.NACHRICHTEN_VORGANG_SERVICE)
	private final VorgangService vorgangService;
	@Qualifier(NachrichtenManagerConfiguration.OZG_CLOUD_FILE_SERVICE_NAME)
	private final OzgCloudFileService ozgCloudFileService;
	@Qualifier(NachrichtenManagerConfiguration.NACHRICHTEN_ATTACHED_ITEM_SERVICE)
	private final AttachedItemService attachedItemService;

	private final RueckfrageMapper rueckfrageMapper;
	private final PostfachNachrichtMapper nachrichtMapper;

	@PostConstruct
	void init() {
		if (isNull(nachrichtenManagerProperties.getAddress())) {
			throw new IllegalStateException("Address of Nachrichten-Manager is not set");
		}
	}

	@PostConstruct
	void logInitalization() {
		LOG.info("Starting Antragraum Service");
	}

	public String getUserNotificationText() {
		return USER_NOTIFICATION_TEMPLATE.formatted(getAntragsraumUrl());
	}

	public String getAntragsraumUrl() {
		return properties.getUrl();
	}

	public Stream<RueckfrageHead> findRueckfragen(String samlToken) {
		verifyToken(samlToken);

		return postfachNachrichtService.findRueckfragen(BAYERN_ID_SERVICE_KONTO_TYPE, getPostfachId(samlToken))
				.map(rueckfrageMapper::fromPostfachNachricht)
				.map(rueckfrage -> enrichByAnswers(rueckfrage, samlToken))
				.map(rueckfrage -> addVorgangData(rueckfrage, vorgangService.getVorgang(rueckfrage.getVorgangId())))
				.map(rueckfrage -> addAccessible(rueckfrage, samlToken));
	}

	RueckfrageHead enrichByAnswers(RueckfrageHead rueckfrage, String samlToken) {
		var rueckfrageBuilder = rueckfrage.toBuilder();

		getFirstSentAt(findAnswers(samlToken, rueckfrage.getId()))
				.ifPresent(firstSentAt -> rueckfrageBuilder.answeredAt(firstSentAt).status(RueckfrageStatus.ANSWERED));

		return rueckfrageBuilder.build();
	}

	Optional<ZonedDateTime> getFirstSentAt(Stream<PostfachNachricht> answers) {
		return answers.sorted(Comparator.comparing(PostfachNachricht::getSentAt)).map(PostfachNachricht::getSentAt).findFirst();
	}

	RueckfrageHead addVorgangData(RueckfrageHead rueckfrage, Vorgang vorgang) {
		return rueckfrage.toBuilder()
				.vorgangName(vorgang.getName())
				.vorgangNummer(vorgang.getNummer())
				.trustLevel(vorgang.getTrustLevel())
				.build();
	}

	RueckfrageHead addAccessible(RueckfrageHead rueckfrage, String samlToken) {
		return rueckfrage.toBuilder().accessible(isAccessible(samlToken, rueckfrage.getTrustLevel())).build();
	}

	public String sendRueckfrageAnswer(String samlToken, String rueckfrageId, PostfachNachricht nachricht) {
		verifyToken(samlToken);
		verifyAccessToPostfach(samlToken, nachricht);

		return postfachNachrichtService.persistAnswer(rueckfrageId, nachricht);
	}

	public Stream<PostfachNachricht> findAnswers(String samlToken, String rueckfrageId) {
		verifyToken(samlToken);

		return postfachNachrichtService.findAnswers(BAYERN_ID_SERVICE_KONTO_TYPE, getPostfachId(samlToken), rueckfrageId);
	}

	public PostfachNachricht getRueckfrage(String samlToken, String id) {
		verifyToken(samlToken);

		var nachricht = nachrichtMapper.fromMapToPostfachMail(postfachNachrichtService.getById(id));

		verifyAccessToPostfach(samlToken, nachricht);

		return nachricht;
	}

	public void getAttachmentContent(AttachmentFileRequest request, OutputStream outputStream) {
		verifyAccessToFile(request);
		ozgCloudFileService.writeFileDataToStream(OzgCloudFileId.from(request.getFileId()), outputStream);
	}

	public OzgCloudFile getAttachmentMetadata(AttachmentFileRequest request) {
		verifyAccessToFile(request);
		return ozgCloudFileService.getFile(OzgCloudFileId.from(request.getFileId()));
	}

	void verifyAccessToFile(AttachmentFileRequest request) {
		verifyToken(request.getSamlToken());
		var nachricht = attachedItemService.getPostfachNachricht(request.getNachrichtId());
		verifyAccessToPostfach(request.getSamlToken(), nachricht);
		verifyFileId(request.getFileId(), nachricht);
	}

	void verifyToken(String token) {
		var errors = verifier.verify(token);
		if (CollectionUtils.isNotEmpty(errors)) {
			throw new SecurityException("SAML Token verification failed. Errors: %s".formatted(errors));
		}
	}

	void verifyAccessToPostfach(String token, PostfachNachricht nachricht) {
		var vorgang = vorgangService.getVorgang(nachricht.getVorgangId());
		verifyPostfachId(token, nachricht, vorgang);
		verifyTrustLevel(token, nachricht, vorgang);
	}

	void verifyPostfachId(String samlToken, PostfachNachricht nachricht, Vorgang vorgang) {
		if (!StringUtils.equals(vorgang.getPostfachId(), getPostfachId(samlToken))) {
			throw new NotFoundException(GenericId.from(nachricht.getId()), POSTFACH_NACHRICHT);
		}
	}

	String getPostfachId(String samlToken) {
		return decrypter.decryptPostfachId(parser.parse(samlToken));
	}

	void verifyTrustLevel(String token, PostfachNachricht nachricht, Vorgang vorgang) {
		if (!isAccessible(token, vorgang.getTrustLevel())) {
			throw new NotFoundException(GenericId.from(nachricht.getId()), POSTFACH_NACHRICHT);
		}
	}

	public boolean isAccessible(String samlToken, String trustLevel) {
		return getTrustLevel(samlToken).getIntValue() >= TrustLevel.fromString(trustLevel).getIntValue();
	}

	TrustLevel getTrustLevel(String samlToken) {
		return TrustLevel.fromString(decrypter.decryptTrustLevel(parseSamlToken(samlToken)));
	}

	Response parseSamlToken(String samlToken) {
		return parser.parse(samlToken);
	}

	void verifyFileId(String fileId, PostfachNachricht nachricht) {
		if (!nachricht.getAttachments().contains(fileId)) {
			throw new NotFoundException(GenericId.from(nachricht.getId()), POSTFACH_NACHRICHT);
		}
	}
}