/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.antragraum;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import jakarta.annotation.PostConstruct;

import org.opensaml.core.xml.schema.XSString;
import org.opensaml.saml.saml2.core.Assertion;
import org.opensaml.saml.saml2.core.Attribute;
import org.opensaml.saml.saml2.core.AttributeStatement;
import org.opensaml.saml.saml2.core.EncryptedAssertion;
import org.opensaml.saml.saml2.core.Response;
import org.opensaml.saml.saml2.encryption.Decrypter;
import org.opensaml.saml.saml2.encryption.EncryptedElementTypeEncryptedKeyResolver;
import org.opensaml.security.credential.Credential;
import org.opensaml.security.credential.CredentialSupport;
import org.opensaml.xmlsec.encryption.support.ChainingEncryptedKeyResolver;
import org.opensaml.xmlsec.encryption.support.EncryptedKeyResolver;
import org.opensaml.xmlsec.encryption.support.InlineEncryptedKeyResolver;
import org.opensaml.xmlsec.encryption.support.SimpleRetrievalMethodEncryptedKeyResolver;
import org.opensaml.xmlsec.keyinfo.impl.CollectionKeyInfoCredentialResolver;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.security.saml2.Saml2Exception;
import org.springframework.stereotype.Service;

import lombok.RequiredArgsConstructor;

@Service
@RequiredArgsConstructor
@ConditionalOnProperty(AntragraumProperties.PROPERTY_ANTRAGSRAUM_URL)
class Saml2Decrypter {
	private final BayernIdSamlConfiguration configuration;

	public static final String LEGACY_POSTKORB_HANDLE_KEY = "urn:oid:2.5.4.18";
	public static final String TRUST_LEVEL_KEY = "urn:oid:1.2.40.0.10.2.1.1.261.94";

	private Decrypter decrypter;
	private static final EncryptedKeyResolver encryptedKeyResolver = new ChainingEncryptedKeyResolver(
			Arrays.asList(new InlineEncryptedKeyResolver(), new EncryptedElementTypeEncryptedKeyResolver(),
					new SimpleRetrievalMethodEncryptedKeyResolver()));

	@PostConstruct
	void init() {
		var credentials = new ArrayList<Credential>();
		var decryptionX509Credential = configuration.getDecryptionCredential();

		var cred = CredentialSupport.getSimpleCredential(decryptionX509Credential.getCertificate(), decryptionX509Credential.getPrivateKey());
		credentials.add(cred);

		var resolver = new CollectionKeyInfoCredentialResolver(credentials);
		var setupDecrypter = new Decrypter(null, resolver, encryptedKeyResolver);
		setupDecrypter.setRootInNewDocument(true);

		decrypter = setupDecrypter;
	}

	public String decryptPostfachId(Response response) {
		decryptResponseElements(response);

		var postfachIdAttribute = getAttributes(response).stream().filter(this::isPostfachId).findFirst();

		return postfachIdAttribute.map(this::getStringValue).orElseThrow();
	}

	boolean isPostfachId(Attribute attribute) {
		return LEGACY_POSTKORB_HANDLE_KEY.equals(attribute.getName());
	}

	public String decryptTrustLevel(Response response) {
		decryptResponseElements(response);

		var trustLevelAttribute = getAttributes(response).stream().filter(this::isTrustLevel).findFirst();

		return trustLevelAttribute.map(this::getStringValue).orElseThrow();
	}

	void decryptResponseElements(Response response) {
		response.getEncryptedAssertions().stream()
				.map(this::decryptAssertion)
				.forEach(assertion -> response.getAssertions().add(assertion));
	}

	private Assertion decryptAssertion(EncryptedAssertion assertion) {
		try {
			return decrypter.decrypt(assertion);
		} catch (Exception ex) {
			throw new Saml2Exception(ex);
		}
	}

	private List<Attribute> getAttributes(Response response) {
		return getAttributeStatement(response).getAttributes();
	}

	private AttributeStatement getAttributeStatement(Response response) {
		return (AttributeStatement) response.getAssertions().getFirst().getStatements().get(1);
	}

	boolean isTrustLevel(Attribute attribute) {
		return TRUST_LEVEL_KEY.equals(attribute.getName());
	}

	private String getStringValue(Attribute attribute) {
		return ((XSString) attribute.getAttributeValues().getFirst()).getValue();
	}
}