/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.email;

import java.util.Objects;

import jakarta.mail.MessagingException;
import jakarta.mail.internet.MimeMessage;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.mail.MailException;
import org.springframework.mail.javamail.JavaMailSender;
import org.springframework.mail.javamail.MimeMessageHelper;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Service;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.email.MailSendRequest.MailAttachment;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
class EMailService {

	static final String HEADER_READ_RECEIPT = "Read-Receipt-To";
	static final String HEADER_CONFIRM_READING = "X-Confirm-reading-to";
	static final String HEADER_DISPOSITION_NOTIFICATION = "Disposition-Notification-To";
	static final String HEADER_SENSITIVITY = "Sensitivity"; // RFC2156
	static final String VALUE_SENSITIVITY = "Company-Confidential";

	@Autowired
	private JavaMailSender sender;

	@Autowired
	private ApplicationEventPublisher publisher;

	@Async
	public void sendEmail(MailSendRequest request) {
		MimeMessage message = sender.createMimeMessage();
		prepareAndSendMail(message, request);
	}

	void prepareAndSendMail(MimeMessage message, MailSendRequest request) {
		MimeMessageHelper helper = buildMultipartHelper(message);

		writeEmail(helper, request);
		addAttachments(helper, request);
		if (request.isReceiptRequired()) {
			requestReceipt(message, request);
		}
		addHeader(message, "Content-Transfer-Encoding", "quoted-printable");
		send(message, request);
	}

	void writeEmail(MimeMessageHelper helper, MailSendRequest request) {
		try {
			var toAddresses = request.getToAddresses().stream().map(this::formatEmailAddress).toArray(String[]::new);
			helper.setTo(toAddresses);
			helper.setFrom(request.getFromAddress());
			helper.setSubject(request.getSubject());
			helper.setText(request.getBody());
			// TODO replyTo ausbauen
			setReplyTo(helper, request);
		} catch (MessagingException e) {
			throw new TechnicalException("Error writing mail", e);
		}
	}

	private String formatEmailAddress(MailRecipient mailRecipient) {
		if (StringUtils.isAllBlank(mailRecipient.firstName(), mailRecipient.lastName())) {
			return mailRecipient.email();
		}
		var builder = new StringBuilder();
		if (StringUtils.isNotEmpty(mailRecipient.firstName())) {
			builder.append(mailRecipient.firstName()).append(" ");
		}
		if (StringUtils.isNotEmpty(mailRecipient.lastName())) {
			builder.append(mailRecipient.lastName()).append(" ");
		}
		return builder.append("<").append(mailRecipient.email()).append(">").toString();
	}

	private void setReplyTo(MimeMessageHelper helper, MailSendRequest request) throws MessagingException {
		if (Objects.nonNull(request.getReplyTo())) {
			helper.setReplyTo(request.getReplyTo());
		}
	}

	void requestReceipt(MimeMessage message, MailSendRequest request) {
		addHeader(message, HEADER_READ_RECEIPT, request.getFromAddress());
		addHeader(message, HEADER_CONFIRM_READING, request.getFromAddress());
		addHeader(message, HEADER_DISPOSITION_NOTIFICATION, request.getFromAddress());
	}

	private void addHeader(MimeMessage message, String header, String value) {
		try {
			message.addHeader(header, value);
		} catch (MessagingException e) {
			throw new TechnicalException(String.format("Error adding header '%s' with value '%s'", header, value), e);
		}
	}

	void addAttachments(MimeMessageHelper helper, MailSendRequest request) {
		request.getAttachments().forEach(attachment -> addAttachment(helper, attachment));
	}

	private void addAttachment(MimeMessageHelper helper, MailAttachment attachment) {
		try {
			if (Objects.isNull(attachment.getContentId())) {
				helper.addAttachment(attachment.getDataSource().getName(), attachment.getDataSource());
			} else {
				helper.addInline(attachment.getContentId(), attachment.getDataSource());

			}

		} catch (MessagingException e) {
			throw new TechnicalException("Error adding Attachment.", e);
		}
	}

	private MimeMessageHelper buildMultipartHelper(MimeMessage message) {
		try {
			return new MimeMessageHelper(message, true);
		} catch (MessagingException e) {
			throw new TechnicalException("Error creating multipart message helper.", e);
		}
	}

	void send(MimeMessage message, MailSendRequest request) {
		try {
			sender.send(message);
			publisher.publishEvent(new MailSentEvent(request));
		} catch (MailException e) {
			LOG.error("Error sending mail", e);
			publisher.publishEvent(new MailSendErrorEvent(request, e.getMessage()));
		}
	}
}
