/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.file;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.net.URLConnection;
import java.util.Base64;
import java.util.Optional;

import jakarta.activation.MimetypesFileTypeMap;

import org.apache.commons.io.IOUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.core.task.TaskExecutor;
import org.springframework.stereotype.Service;

import de.ozgcloud.apilib.file.OzgCloudFileService;
import de.ozgcloud.apilib.vorgang.OzgCloudFileIdMapper;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.NachrichtenManagerConfiguration;
import de.ozgcloud.nachrichten.postfach.FileId;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Service(NachrichtenManagerConfiguration.ATTACHMENT_FILE_SERVICE_NAME) // NOSONAR
@RequiredArgsConstructor
@Log4j2
public class AttachmentFileService {

	private static final int BUFFER_SIZE = 4 * 1024;
	static final String ATTACHMENT_NAME = "PostfachAttachment";

	@Qualifier(NachrichtenManagerConfiguration.OZG_CLOUD_FILE_SERVICE_NAME) // NOSONAR
	private final OzgCloudFileService ozgCloudFileService;
	private final OzgCloudFileIdMapper fileIdMapper;
	@Qualifier(NachrichtenManagerConfiguration.ATTACHMENT_FILE_MAPPER_NAME) // NOSONAR
	private final AttachmentFileMapper attachmentFileMapper;

	private final TaskExecutor taskExecutor;

	private final MimetypesFileTypeMap mimetypesFileTypeMap;

	public AttachmentFile getFile(FileId fileId) {
		var ozgCloudFile = ozgCloudFileService.getFile(fileIdMapper.toFileId(fileId.toString()));
		return attachmentFileMapper.fromOzgCloudFile(ozgCloudFile);
	}

	public InputStream getFileContent(FileId fileId) {
		var inputStream = createInputStream();
		writeFileContent(fileId.toString(), connectToOutputStream(inputStream));
		return inputStream;
	}

	PipedInputStream createInputStream() {
		return new PipedInputStream(BUFFER_SIZE);
	}

	OutputStream connectToOutputStream(PipedInputStream inputStream) {
		try {
			return new PipedOutputStream(inputStream);
		} catch (IOException e) {
			throw new TechnicalException("Cannot read file content", e);
		}
	}

	void writeFileContent(String fileId, OutputStream outputStream) {
		taskExecutor.execute(() -> {
			ozgCloudFileService.writeFileDataToStream(fileIdMapper.toFileId(fileId), outputStream);
			IOUtils.closeQuietly(outputStream, e -> LOG.warn("Cannot close output stream", e));
		});
	}

	public String uploadAttachmentFile(AttachmentFile attachmentFile, String fileContent) {
		var decodedContent = Base64.getDecoder().decode(fileContent);
		var ozgCloudFileId = ozgCloudFileService.uploadFile(
				attachmentFileMapper.toOzgCloudUploadFile(addContentTypeIfMissing(attachmentFile, decodedContent)),
				toInputStream(decodedContent));
		return ozgCloudFileId.toString();
	}

	AttachmentFile addContentTypeIfMissing(AttachmentFile attachmentFile, byte[] fileContent) {
		if (StringUtils.isBlank(attachmentFile.getContentType())) {
			return attachmentFile.toBuilder().contentType(getContentType(attachmentFile, fileContent)).build();
		}
		return attachmentFile;
	}

	String getContentType(AttachmentFile attachmentFile, byte[] fileContent) {
		return getTypeByFileName(attachmentFile).or(() -> getTypeByContent(fileContent)).orElseGet(() -> getByMimeTypes(attachmentFile));
	}

	Optional<String> getTypeByFileName(AttachmentFile attachmentFile) {
		var contentType = URLConnection.getFileNameMap().getContentTypeFor(attachmentFile.getName());
		return Optional.ofNullable(contentType);
	}

	Optional<String> getTypeByContent(byte[] fileContent) {
		try (var contentStream = toInputStream(fileContent)) {
			return Optional.ofNullable(URLConnection.guessContentTypeFromStream(contentStream));
		} catch (IOException e) {
			LOG.warn("IO-Exception while guessing content type", e);
		}
		return Optional.empty();
	}

	String getByMimeTypes(AttachmentFile attachmentFile) {
		return mimetypesFileTypeMap.getContentType(attachmentFile.getName());
	}

	InputStream toInputStream(byte[] content) {
		return new ByteArrayInputStream(content);
	}
}
