/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static java.util.Objects.*;

import java.util.Iterator;

import de.ozgcloud.vorgang.common.GrpcQueryOperator;
import de.ozgcloud.vorgang.vorgang.GrpcFindVorgangRequest;
import de.ozgcloud.vorgang.vorgang.GrpcQuery;
import de.ozgcloud.vorgang.vorgang.GrpcVorgangHeader;
import de.ozgcloud.vorgang.vorgang.GrpcVorgangQueryExpression;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import lombok.RequiredArgsConstructor;

@RequiredArgsConstructor
public class FindVorgangIterator implements Iterator<GrpcVorgangHeader> {

	static final int LIMIT = 50;
	static final String PATH_SERVICEKONTO_TYPE = "Vorgang.header.serviceKonto.type";
	static final String PATH_POSTFACH_ID = "Vorgang.header.serviceKonto.postfachAddresses.identifier.postfachId";

	private final VorgangServiceBlockingStub vorgangServiceStub;
	private final String servicekontoType;
	private final String postfachId;

	private Iterator<GrpcVorgangHeader> vorgangHeaderIterator;
	private int lastVorgangListSize = 0;
	private int offset = 0;

	@Override
	public boolean hasNext() {
		if (shouldRequestNext()) {
			offset += lastVorgangListSize;
			var findVorgangResponse = vorgangServiceStub.findVorgang(buildFindRequest(offset));
			vorgangHeaderIterator = findVorgangResponse.getVorgangList().iterator();
			lastVorgangListSize = findVorgangResponse.getVorgangList().size();
		}
		return vorgangHeaderIterator.hasNext();
	}

	boolean shouldRequestNext() {
		return isNull(vorgangHeaderIterator) || (!vorgangHeaderIterator.hasNext() && lastVorgangListSize == LIMIT);
	}

	@Override
	public GrpcVorgangHeader next() {
		return vorgangHeaderIterator.next();
	}

	GrpcFindVorgangRequest buildFindRequest(int offset) {
		return GrpcFindVorgangRequest.newBuilder()
				.setOffset(offset)
				.setLimit(LIMIT)
				.setQuery(buildQuery())
				.build();
	}

	GrpcQuery buildQuery() {
		return GrpcQuery.newBuilder()
				.addExpressions(buildServicekontoExpression())
				.addExpressions(buildPostfachIdExpression())
				.build();
	}

	GrpcVorgangQueryExpression buildServicekontoExpression() {
		return GrpcVorgangQueryExpression.newBuilder()
				.setPath(PATH_SERVICEKONTO_TYPE)
				.setOperator(GrpcQueryOperator.EQUAL)
				.setOperandStringValue(servicekontoType)
				.build();
	}

	GrpcVorgangQueryExpression buildPostfachIdExpression() {
		return GrpcVorgangQueryExpression.newBuilder()
				.setPath(PATH_POSTFACH_ID)
				.setOperator(GrpcQueryOperator.EQUAL)
				.setOperandStringValue(postfachId)
				.build();
	}
}
