/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static java.util.Objects.*;

import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.context.annotation.Primary;
import org.springframework.stereotype.Service;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.user.OzgCloudUserId;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;
import de.ozgcloud.nachrichten.NachrichtenManagerConfiguration;
import de.ozgcloud.nachrichten.attributes.ClientAttributeService;
import de.ozgcloud.nachrichten.common.vorgang.VorgangService;
import de.ozgcloud.nachrichten.file.AttachmentFile;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.Direction;

@Service
@Primary
public class PersistPostfachNachrichtServiceImpl implements PersistPostfachNachrichtService {

	private static final String NOT_IMPLEMENTED_MESSAGE = "Not implemented. Use gRPC API instead.";

	static final String CLIENT = "OzgCloud_NachrichtenManager";
	static final String ITEM_NAME = "PostfachMail";

	static final String CREATE_ATTACHED_ITEM_ORDER = "CREATE_ATTACHED_ITEM";
	static final String UPDATE_ATTACHED_ITEM_ORDER = "UPDATE_ATTACHED_ITEM";

	static final String CLIENT_FIELD = "client";
	static final String VORGANG_ID_FIELD = "vorgangId";
	static final String ITEM_NAME_FIELD = "itemName";
	static final String ITEM_FIELD = "item";

	@Autowired
	@Qualifier(NachrichtenManagerConfiguration.NACHRICHTEN_VORGANG_SERVICE)
	private VorgangService vorgangService;

	@Autowired
	@Qualifier(NachrichtenManagerConfiguration.OZG_CLOUD_COMMAND_SERVICE_NAME)
	private OzgCloudCommandService commandService;

	@Autowired
	private AttachedItemRemoteService attachedItemRemoteService;
	@Autowired
	private PostfachNachrichtMapper postfachNachrichtMapper;
	@Autowired
	private ClientAttributeService clientAttributeService;

	@Autowired(required = false)
	@Qualifier("persistPostfachNachrichtByCommandService")
	private PersistPostfachNachrichtService vorgangManagerNachrichtService;

	@Override
	public void persistNachricht(Optional<String> userId, PostfachNachricht nachricht) {
		commandService.create(buildNachrichtCommand(userId, nachricht));

		setHasNewPostfachNachricht(nachricht);
	}

	OzgCloudCommand buildNachrichtCommand(Optional<String> userId, PostfachNachricht nachricht) {
		return StringUtils.isBlank(nachricht.getId()) ? buildCreateCommand(userId, nachricht) : buildUpdateCommand(userId, nachricht);
	}

	OzgCloudCommand buildUpdateCommand(Optional<String> userId, PostfachNachricht nachricht) {
		return createBaseCommandBuilder(userId, nachricht)
				.order(UPDATE_ATTACHED_ITEM_ORDER)
				.relationId(NachrichtId.from(nachricht.getId()))
				.bodyObject(buildUpdateItem(nachricht))
				.build();
	}

	private Map<String, Object> buildUpdateItem(PostfachNachricht nachricht) {
		return Map.of(
				CLIENT_FIELD, CLIENT,
				ITEM_FIELD, postfachNachrichtMapper.mapToMap(nachricht));
	}

	// TODO pruefen ob referencedNachricht existiert
	// TODO pruefen ob referencedNachricht in dem postfach des anfragenden liegt
	@Override
	public String persistAnswer(String referencedNachrichtId, PostfachNachricht answer) {
		var enhancedAnswer = answer.toBuilder().referencedNachricht(referencedNachrichtId).build();
		var command = buildCreateCommand(Optional.ofNullable(answer.getCreatedBy()), enhancedAnswer);

		var commandId = commandService.create(command).getId().toString();

		setHasNewPostfachNachricht(answer);

		return commandId;
	}

	OzgCloudCommand buildCreateCommand(Optional<String> userId, PostfachNachricht nachricht) {
		return createBaseCommandBuilder(userId, nachricht)
				.order(CREATE_ATTACHED_ITEM_ORDER)
				.relationId(OzgCloudVorgangId.from(nachricht.getVorgangId()))
				.bodyObject(buildCreateItem(nachricht))
				.build();
	}

	OzgCloudCommand.OzgCloudCommandBuilder createBaseCommandBuilder(Optional<String> userId, PostfachNachricht nachricht) {
		var commandBuilder = OzgCloudCommand.builder().vorgangId(OzgCloudVorgangId.from(nachricht.getVorgangId()));

		userId.map(OzgCloudUserId::from).ifPresent(commandBuilder::createdBy);

		return commandBuilder;
	}

	Map<String, Object> buildCreateItem(PostfachNachricht nachricht) {
		return Map.of(
				CLIENT_FIELD, CLIENT,
				VORGANG_ID_FIELD, nachricht.getVorgangId(),
				ITEM_NAME_FIELD, ITEM_NAME,
				ITEM_FIELD, postfachNachrichtMapper.mapToMap(nachricht));
	}

	void setHasNewPostfachNachricht(PostfachNachricht nachricht) {
		if (isIncomingNachricht(nachricht)) {
			clientAttributeService.setHasNewPostfachNachricht(nachricht.getVorgangId());
		}
	}

	private boolean isIncomingNachricht(PostfachNachricht nachricht) {
		return nachricht.getDirection() == Direction.IN;
	}

	@Override
	public Stream<Map<String, Object>> findByVorgangAsMap(String vorgangId) {
		if (vorgangManagerNachrichtService != null) {
			return vorgangManagerNachrichtService.findByVorgangAsMap(vorgangId);
		}
		throw new UnsupportedOperationException(NOT_IMPLEMENTED_MESSAGE);
	}

	@Override
	public Optional<Map<String, Object>> findById(String nachrichtId) {
		if (vorgangManagerNachrichtService != null) {
			return vorgangManagerNachrichtService.findById(nachrichtId);
		}
		throw new UnsupportedOperationException(NOT_IMPLEMENTED_MESSAGE);
	}

	@Override
	public void patch(String postfachMailId, Map<String, Object> propertyMap) {
		if (nonNull(vorgangManagerNachrichtService)) {
			vorgangManagerNachrichtService.patch(postfachMailId, propertyMap);
			return;
		}
		throw new UnsupportedOperationException(NOT_IMPLEMENTED_MESSAGE);
	}

	@Override
	public Map<String, Object> getById(String itemId) {
		if (vorgangManagerNachrichtService != null) {
			return vorgangManagerNachrichtService.getById(itemId);
		}
		throw new UnsupportedOperationException(NOT_IMPLEMENTED_MESSAGE);
	}

	@Override
	public String persistAttachment(String vorgangId, AttachmentFile attachment) {
		if (nonNull(vorgangManagerNachrichtService)) {
			return vorgangManagerNachrichtService.persistAttachment(vorgangId, attachment);
		}
		throw new UnsupportedOperationException(NOT_IMPLEMENTED_MESSAGE);
	}

	@Override
	public Stream<PostfachNachricht> findRueckfragen(String servicekontoType, String postfachId) {
		return vorgangService.findVorgangIds(servicekontoType, postfachId)
				.flatMap(attachedItemRemoteService::findReplyAllowedPostfachNachrichts);
	}

	@Override
	public Stream<PostfachNachricht> findAnswers(String serviceKontoType, String postfachId, String rueckfrageId) {
		return findPostfachNachricht(serviceKontoType, postfachId)
				.filter(postfachNachricht -> hasMatchingReferenceNachrichtId(postfachNachricht, rueckfrageId));
	}

	private Stream<PostfachNachricht> findPostfachNachricht(String serviceKontoType, String postfachId) {
		return vorgangService.findVorgangIds(serviceKontoType, postfachId).flatMap(attachedItemRemoteService::findByVorgangId);
	}

	private boolean hasMatchingReferenceNachrichtId(PostfachNachricht postfachNachricht, String postfachNachrichtId) {
		return StringUtils.equals(postfachNachricht.getReferencedNachricht(), postfachNachrichtId);
	}
}