/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Predicate;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.event.EventListener;
import org.springframework.stereotype.Component;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.PostfachNachrichtReceivedEvent;
import de.ozgcloud.command.VorgangAttachedItemCreatedEvent;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.Direction;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Component
@Log4j2
@RequiredArgsConstructor
public class PostfachEventListener {

	private static final String IS_SEND_POSTFACH_NACHRICHT_COMMAND = "{T(de.ozgcloud.nachrichten.postfach.PostfachEventListener).IS_SEND_POSTFACH_NACHRICHT.test(event.getSource())}";
	public static final Predicate<Command> IS_SEND_POSTFACH_NACHRICHT = command -> StringUtils.equals(command.getOrder(), "SEND_POSTFACH_NACHRICHT");
	public static final String KEY_ITEM_NAME = "itemName";
	public static final Predicate<Command> IS_INCOMING_POSTFACH_NACHRICHT_CREATED = command ->
			StringUtils.equals(command.getOrder(), "CREATE_ATTACHED_ITEM")
			&& StringUtils.equals(MapUtils.getString(command.getBodyObject(), KEY_ITEM_NAME), AttachedItemRemoteService.ITEM_NAME)
			&& StringUtils.equals(MapUtils.getString(command.getBodyObject(), PostfachNachricht.FIELD_DIRECTION), "IN");
	private static final String IS_INCOMING_POSTFACH_NACHRICHT_ITEM_CREATED_CONDITION =
			"{T(de.ozgcloud.nachrichten.postfach.PostfachEventListener).IS_INCOMING_POSTFACH_NACHRICHT_CREATED.test(event.getCommand())}";
	private final PostfachService service;
	private final ApplicationEventPublisher publisher;

	@EventListener(condition = IS_SEND_POSTFACH_NACHRICHT_COMMAND)
	public void sendPostfachNachricht(CommandCreatedEvent event) {
		var command = event.getSource();
		try {
			service.sendMail(command.getId(), command.getCreatedBy(), buildNachricht(command, command.getBodyObject()));
		} catch (RuntimeException e) {
			service.publishMailSentFailedEvent(command.getId(), "Error on sending Postfach Nachricht: " + e.getMessage());
			LOG.error("Error on sending Postfach Nachricht.", e);
		}
	}

	private PostfachNachricht buildNachricht(Command command, Map<String, Object> commandBody) {
		var postfachMailBuilder = PostfachNachricht.builder()
				.id(isNewNachricht(command) ? null : command.getRelationId())
				.vorgangId(command.getVorgangId())
				.replyOption(PostfachNachricht.ReplyOption.valueOf(MapUtils.getString(commandBody, PostfachNachricht.FIELD_REPLY_OPTION)))
				.createdBy(command.getCreatedBy())
				.direction(Direction.OUT)
				.subject(MapUtils.getString(commandBody, PostfachNachricht.FIELD_SUBJECT))
				.mailBody(MapUtils.getString(commandBody, PostfachNachricht.FIELD_MAIL_BODY))
				.attachments(getStringListValue(commandBody, PostfachNachricht.FIELD_ATTACHMENTS));

		Optional.ofNullable(getPostfachAddress(commandBody)).map(this::buildPostfachAddress).ifPresent(postfachMailBuilder::postfachAddress);

		return postfachMailBuilder.build();
	}

	private boolean isNewNachricht(Command command) {
		return Objects.isNull(command.getRelationId()) || StringUtils.equals(command.getVorgangId(), command.getRelationId());
	}

	private PostfachAddress buildPostfachAddress(Map<String, Object> postfachAddress) {
		return PostfachAddress.builder()
				.type(MapUtils.getIntValue(postfachAddress, PostfachAddress.TYPE_FIELD))
				.version(MapUtils.getString(postfachAddress, PostfachAddress.VERSION_FIELD))
				.identifier(buildIdentifier(postfachAddress))
				.serviceKontoType(MapUtils.getString(postfachAddress, PostfachAddress.SERVICEKONTO_TYPE_FIELD))
				.build();
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> getPostfachAddress(Map<String, Object> commandBody) {
		return (Map<String, Object>) commandBody.get(PostfachNachricht.POSTFACH_ADDRESS_FIELD);
	}

	private PostfachAddressIdentifier buildIdentifier(Map<String, Object> postfachAddress) {
		var identifier = getIdentifier(postfachAddress);
		return StringBasedIdentifier.builder().postfachId(MapUtils.getString(identifier, PostfachAddress.FIELD_POSTFACH_ID)).build();
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> getIdentifier(Map<String, Object> postfachAddress) {
		return (Map<String, Object>) postfachAddress.get(PostfachAddress.IDENTIFIER_FIELD);
	}

	List<String> getStringListValue(Map<String, Object> commandBody, String fieldName) {
		return Optional.ofNullable(commandBody.get(fieldName))
				.map(this::mapObjectToList)
				.orElse(Collections.emptyList());
	}

	private List<String> mapObjectToList(Object object) {
		if (object instanceof Collection<?> objects) {
			return objects.stream().map(Objects::toString).toList();
		} else {
			return List.of(object.toString());
		}
	}

	@EventListener(condition = IS_INCOMING_POSTFACH_NACHRICHT_ITEM_CREATED_CONDITION)
	public void onIncomingPostfachNachrichtItemCreated(VorgangAttachedItemCreatedEvent event) {
		publisher.publishEvent(new PostfachNachrichtReceivedEvent(event.getCommand(), event.getCreatedResource()));
	}
}