/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid;

import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.BiFunction;

import org.apache.commons.io.IOUtils;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageRequest;
import io.grpc.stub.CallStreamObserver;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
class FileSender {

	static final int CHUNK_SIZE = 4 * 1024;

	private final AtomicBoolean metaDataSent = new AtomicBoolean(false);
	private final AtomicBoolean done = new AtomicBoolean(false);

	private final BiFunction<byte[], Integer, GrpcSendBayernIdMessageRequest> chunkBuilder;
	private final CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;
	private final GrpcSendBayernIdMessageRequest metadata;
	private final StreamReader streamReader;

	public FileSender(BiFunction<byte[], Integer, GrpcSendBayernIdMessageRequest> chunkBuilder,
			CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver, GrpcSendBayernIdMessageRequest metadata,
			InputStream inputStream) {
		this.chunkBuilder = chunkBuilder;
		this.requestObserver = requestObserver;
		this.metadata = metadata;
		this.streamReader = new StreamReader(inputStream);
	}

	public void send() {
		if (!done.get()) {
			sendMetadata();
			while (!done.get() && requestObserver.isReady()) {
				LOG.debug("Sending next chunk.");
				sendNextChunk();
			}
			LOG.debug("Finished or waiting to become ready.");
		}
	}

	void sendMetadata() {
		if (metaDataSent.getAndSet(true)) {
			return;
		}
		LOG.debug("Sending Attachment Metadata.");
		requestObserver.onNext(metadata);
	}

	long sendNextChunk() {
		byte[] contentToSend = streamReader.getNextData();

		if (streamReader.getLastReadSize() > 0) {
			sendChunk(contentToSend, streamReader.getLastReadSize());
		} else {
			endTransfer();
		}
		return contentToSend.length;
	}

	void endTransfer() {
		done.set(true);
		LOG.debug("File Transfer done. Closing stream.");
		streamReader.close();
	}

	void sendChunk(byte[] content, int length) {
		LOG.debug("Sending {} byte Data.", length);
		var chunk = chunkBuilder.apply(content, length);
		requestObserver.onNext(chunk);
	}

	public boolean isDone() {
		return done.get();
	}

	@RequiredArgsConstructor
	class StreamReader {

		private final InputStream inStream;
		private final byte[] buffer = new byte[CHUNK_SIZE];
		@Getter
		private int lastReadSize = 0;
		@Getter
		private final AtomicBoolean done = new AtomicBoolean(false);

		byte[] getNextData() {
			readNext();
			return buffer;
		}

		void close() {
			IOUtils.closeQuietly(inStream);
		}

		void readNext() {
			try {
				lastReadSize = inStream.read(buffer, 0, CHUNK_SIZE);
			} catch (IOException e) {
				throw new TechnicalException("Error on reading a single chunk", e);
			}
		}
	}

}
