/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid;

import static java.util.Objects.*;

import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.BiFunction;
import java.util.function.Function;

import de.ozgcloud.common.binaryfile.BinaryFileUploadStreamObserver;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageRequest;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageResponse;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.Builder;
import lombok.Getter;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Builder
public class MessageWithFilesSender {

	static final int CHUNK_SIZE = 4 * 1024;

	@Getter
	private final CompletableFuture<GrpcSendBayernIdMessageResponse> resultFuture = new CompletableFuture<>();
	private final AtomicBoolean done = new AtomicBoolean(false);
	private final AtomicBoolean metaDataSent = new AtomicBoolean(false);

	private final Function<StreamObserver<GrpcSendBayernIdMessageResponse>, CallStreamObserver<GrpcSendBayernIdMessageRequest>> reqObserverBuilder;
	private final GrpcSendBayernIdMessageRequest messageMetadata;
	@Builder.Default
	private final List<String> attachmentIds = Collections.emptyList();
	private final Function<String, BayernIdAttachment> toAttachment;
	private final Function<BayernIdAttachment, GrpcSendBayernIdMessageRequest> attachmentMetadataMapper;
	private final BiFunction<byte[], Integer, GrpcSendBayernIdMessageRequest> chunkBuilder;

	private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;
	private int currentAttachmentIndex = 0;
	private FileSender currentFileSender;

	public MessageWithFilesSender send() {
		var responseStreamObserver = BinaryFileUploadStreamObserver.create(resultFuture, this::sendNext);
		requestObserver = reqObserverBuilder.apply(responseStreamObserver);
		return this;
	}

	public void sendNext() {
		if (done.get()) {
			return;
		}
		waitForObserver();
		sendMetadata();
		sendAttachments();
	}

	synchronized void waitForObserver() {
		while (isNull(requestObserver)) {
			try {
				LOG.debug("wait for observer");
				wait(300);
			} catch (InterruptedException e) {
				LOG.error("Error on waiting for request Observer.", e);
				Thread.currentThread().interrupt();
			}
		}
	}

	void sendMetadata() {
		if (metaDataSent.getAndSet(true)) {
			return;
		}
		requestObserver.onNext(messageMetadata);
	}

	void sendAttachments() {
		if (attachmentIds.isEmpty()) {
			completeRequest();
		} else {
			sendData();
		}
	}

	void sendData() {
		Optional<FileSender> fileSender;
		while ((fileSender = getFileSender()).isPresent() && requestObserver.isReady()) {
			fileSender.get().send();
		}
		if (fileSender.isEmpty()) {
			completeRequest();
		}
	}

	Optional<FileSender> getFileSender() {
		if (isNull(currentFileSender)) {
			currentFileSender = buildNextFileSender();
			return Optional.of(currentFileSender);
		}
		if (!currentFileSender.isDone()) {
			return Optional.of(currentFileSender);
		}
		if (currentAttachmentIndex < attachmentIds.size()) {
			currentFileSender = buildNextFileSender();
			return Optional.of(currentFileSender);
		}
		return Optional.empty();
	}

	FileSender buildNextFileSender() {
		var attachment = toAttachment.apply(attachmentIds.get(currentAttachmentIndex++));
		return new FileSender(chunkBuilder, requestObserver, attachmentMetadataMapper.apply(attachment), attachment.getContent());
	}

	void completeRequest() {
		done.set(true);
		requestObserver.onCompleted();
	}

	public void cancelOnTimeout() {
		LOG.warn("File transfer canceled on timeout");
		resultFuture.cancel(true);
		requestObserver.onError(new TechnicalException("Timeout on waiting for upload."));
	}

	public void cancelOnError(Throwable t) {
		resultFuture.cancel(true);
		requestObserver.onError(t);
	}
}
