/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.osi;

import java.util.Arrays;
import java.util.Objects;
import java.util.function.Supplier;
import java.util.stream.Stream;

import jakarta.annotation.PostConstruct;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpMethod;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Service;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.UriComponentsBuilder;

import de.ozgcloud.nachrichten.postfach.NotConfiguredException;
import de.ozgcloud.nachrichten.postfach.PostfachBadRequestException;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachRemoteService;
import de.ozgcloud.nachrichten.postfach.PostfachRuntimeException;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
@ConditionalOnProperty(prefix = OsiPostfachProperties.PROXY_API_PREFIX, name = { "url", "key", "realm" })
class OsiPostfachRemoteService implements PostfachRemoteService {

	static final String POSTFACH_TYPE = "OSI";

	@Autowired
	private OsiPostfachMessageMapper mapper;
	@Autowired
	private OsiPostfachProperties properties;
	@Autowired(required = false)
	private RestTemplate restTemplate;

	@PostConstruct
	public void init() {
		LOG.info("OSI Postfach remote service initialized.");
		if (isNotConfigured()) {
			LOG.warn("OSI Postfach Api not configured. Failed to read config " + OsiPostfachProperties.PROXY_API_PREFIX);
		}
	}

	@Override
	public void sendMessage(PostfachNachricht nachricht) {
		checkWhetherIsConfigured();

		HttpEntity<Message> request = new HttpEntity<>(mapper.toOsiMessage(nachricht));

		var response = executeHandlingException(
				() -> restTemplate.exchange(properties.getUrl(), HttpMethod.POST, request, String.class));

		sentFailed(response);
	}

	@Override
	public Stream<PostfachNachricht> getAllMessages() {
		checkWhetherIsConfigured();

		ResponseEntity<Message[]> response = executeHandlingException(
				() -> restTemplate.exchange(properties.getUrl(), HttpMethod.GET, null, Message[].class));

		if (Objects.isNull(response.getBody())) {
			LOG.warn("OSI Postfach response with an empty body");
			return Stream.empty();
		}
		return Arrays.stream(response.getBody()).map(mapper::toPostfachNachricht);
	}

	private void sentFailed(ResponseEntity<String> response) {
		if (isSentFailed(response)) {
			throw new OsiPostfachServerProcessException();
		}
	}

	private boolean isSentFailed(ResponseEntity<String> response) {
		return StringUtils.equals(response.getBody(), String.valueOf(false));
	}

	@Override
	public void deleteMessage(String messageId) {
		checkWhetherIsConfigured();

		UriComponentsBuilder builder = UriComponentsBuilder.fromHttpUrl(properties.getUrl()).queryParam("messageGuid", messageId);
		executeHandlingException(() -> restTemplate.exchange(builder.toUriString(), HttpMethod.DELETE, null, String.class));
	}

	private void checkWhetherIsConfigured() {
		if (isNotConfigured()) {
			throw new NotConfiguredException(
					"Postfach %s is not completely configured. Sending and receiving of postfach mails is not possible.".formatted(POSTFACH_TYPE));
		}
	}

	private boolean isNotConfigured() {
		return Objects.isNull(restTemplate) || Objects.isNull(properties.getUrl()) || Objects.isNull(properties.getKey());
	}

	private <T> T executeHandlingException(Supplier<T> runnable) {
		try {
			return runnable.get();
		} catch (HttpClientErrorException e) {
			throw new PostfachBadRequestException(e);
		} catch (RuntimeException e) {
			throw new PostfachRuntimeException(e);
		}
	}

	@Override
	public String getPostfachType() {
		return POSTFACH_TYPE;
	}

	@Override
	public boolean isReplyAllowed() {
		return true;
	}

}