/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.antragraum;

import static org.assertj.core.api.Assertions.*;

import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoField;
import java.time.temporal.ChronoUnit;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;

import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.Direction;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.ReplyOption;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;

class AntragraumNachrichtMapperTest {

	private final AntragraumNachrichtMapper mapper = Mappers.getMapper(AntragraumNachrichtMapper.class);

	@Nested
	class TestMapRueckfrage {

		@Test
		void shouldMapVorgangId() {
			var result = toGrpc();

			assertThat(result.getVorgangId()).isEqualTo(RueckfrageHeadTestFactory.VORGANG_ID);
		}

		@Test
		void shouldMapId() {
			var result = toGrpc();

			assertThat(result.getId()).isEqualTo(RueckfrageHeadTestFactory.ID);
		}

		@Test
		void shouldMapVorgangName() {
			var result = toGrpc();

			assertThat(result.getVorgangName()).isEqualTo(RueckfrageHeadTestFactory.VORGANG_NAME);
		}

		@Test
		void shouldMapStatus() {
			var result = toGrpc();

			assertThat(result.getStatus()).isEqualTo(RueckfrageStatus.NEW.name());
		}

		@Test
		void shouldMapSentAt() {
			var result = toGrpc();

			assertThat(result.getSentAt()).isEqualTo(PostfachNachrichtTestFactory.SENT_AT.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME));
		}

		private GrpcRueckfrageHead toGrpc() {
			return mapper.toGrpc(RueckfrageHeadTestFactory.create());
		}

	}

	@Nested
	class TestMapAnswerToPostfachNachricht {
		@Test
		void shouldMapText() {
			var result = fromRueckfrageAnswer();

			assertThat(result.getMailBody()).isEqualTo(GrpcRueckfrageAnswerTestFactory.TEXT);
		}

		@Test
		void shouldMapMessageId() {
			var result = fromRueckfrageAnswer();

			assertThat(result.getMessageId()).isEqualTo(GrpcRueckfrageAnswerTestFactory.RUECKFRAGE_ID);
		}

		@Test
		void shouldMapAttachmentIds() {
			var result = fromRueckfrageAnswer();

			assertThat(result.getAttachments()).isEqualTo(GrpcRueckfrageAnswerTestFactory.ATTACHMENT_ID_LIST);
		}

		@Test
		void shouldSetDirection() {
			var result = fromRueckfrageAnswer();

			assertThat(result.getDirection()).isEqualTo(Direction.IN);
		}

		@Test
		void shouldSetReplyOption() {
			var result = fromRueckfrageAnswer();

			assertThat(result.getReplyOption()).isEqualTo(ReplyOption.FORBIDDEN);
		}

		@Test
		void shouldSetCreateAt() {
			var result = fromRueckfrageAnswer();

			assertThat(result.getCreatedAt()).isCloseTo(ZonedDateTime.now(), within(2, ChronoUnit.SECONDS));
		}

		@Test
		void shouldSetSentAt() {
			var result = fromRueckfrageAnswer();

			assertThat(result.getSentAt()).isCloseTo(PostfachNachrichtTestFactory.SENT_AT, within(2, ChronoUnit.SECONDS));
		}

		@DisplayName("should set sentAt to now without milliseconds if value is missing")
		@Test
		void shouldSetSentAtIfMissing() {
			var rueckfrage = GrpcRueckfrageAnswerTestFactory.createBuilder().setSentAt(StringUtils.EMPTY).build();

			var result = mapper.fromRueckfrageAnswer(rueckfrage);

			assertThat(result.getSentAt()).isCloseTo(ZonedDateTime.now(), within(2, ChronoUnit.SECONDS));
			assertThat(result.getSentAt().get(ChronoField.MILLI_OF_SECOND)).isZero();
		}

		private PostfachNachricht fromRueckfrageAnswer() {
			return mapper.fromRueckfrageAnswer(GrpcRueckfrageAnswerTestFactory.create());
		}
	}
}