/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.antragraum;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.OutputStream;
import java.time.ZonedDateTime;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.opensaml.saml.saml2.core.Response;

import de.ozgcloud.apilib.common.errorhandling.NotFoundException;
import de.ozgcloud.apilib.file.OzgCloudFile;
import de.ozgcloud.apilib.file.OzgCloudFileId;
import de.ozgcloud.apilib.file.OzgCloudFileService;
import de.ozgcloud.apilib.file.OzgCloudFileTestFactory;
import de.ozgcloud.nachrichten.NachrichtenManagerProperties;
import de.ozgcloud.nachrichten.common.vorgang.GrpcPostfachAddressTestFactory;
import de.ozgcloud.nachrichten.common.vorgang.GrpcServiceKontoTestFactory;
import de.ozgcloud.nachrichten.common.vorgang.Vorgang;
import de.ozgcloud.nachrichten.common.vorgang.VorgangService;
import de.ozgcloud.nachrichten.common.vorgang.VorgangTestFactory;
import de.ozgcloud.nachrichten.postfach.AttachedItemService;
import de.ozgcloud.nachrichten.postfach.PersistPostfachNachrichtService;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtMapper;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;
import lombok.SneakyThrows;

class AntragraumServiceTest {

	@Spy
	@InjectMocks
	private AntragraumService service;

	@Mock
	private AntragraumProperties properties;
	@Mock
	private NachrichtenManagerProperties nachrichtenManagerProperties;
	@Mock
	private PersistPostfachNachrichtService postfachNachrichtService;
	@Mock
	private Saml2Verifier verifier;
	@Mock
	private Saml2Parser parser;
	@Mock
	private Saml2Decrypter decrypter;
	@Mock
	private RueckfrageMapper rueckfrageMapper;
	@Mock
	private VorgangService vorgangService;
	@Mock
	private PostfachNachrichtMapper nachrichtMapper;
	@Mock
	private AttachedItemService attachedItemService;
	@Mock
	private OzgCloudFileService ozgCloudFileService;

	@Nested
	class TestGetAntragsraumUrl {

		private static final String URL = "url";

		@Test
		void shouldReturnUrl() {
			when(properties.getUrl()).thenReturn(URL);

			var url = service.getAntragsraumUrl();

			assertThat(url).isEqualTo(URL);
		}
	}

	@DisplayName("Find rueckfragen")
	@Nested
	class TestFindRueckfragen {

		@Mock
		private Response response;

		private static final String SAML_TOKEN = "TOKEN";
		private static final String POSTFACH_ID = "dummyPostfachId";

		private final RueckfrageHead rueckfrage = RueckfrageHeadTestFactory.create();
		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();

		private final Vorgang vorgang = VorgangTestFactory.create();

		@BeforeEach
		void mock() {
			when(postfachNachrichtService.findRueckfragen(any(), any())).thenReturn(Stream.of(postfachNachricht));

			when(rueckfrageMapper.fromPostfachNachricht(any())).thenReturn(rueckfrage);
			doReturn(rueckfrage).when(service).enrichByAnswers(any(), any());
			when(vorgangService.getVorgang(any())).thenReturn(vorgang);
			doReturn(POSTFACH_ID).when(service).getPostfachId(any());

			doReturn(rueckfrage).when(service).addVorgangData(any(), any());
			doReturn(rueckfrage).when(service).addAccessible(any(), any());

			doNothing().when(service).verifyToken(any());
		}

		@Test
		void shouldCallVerify() {
			findRueckfragen();

			verify(service).verifyToken(SAML_TOKEN);
		}

		@Test
		void shouldGetPostfachId() {
			findRueckfragen();

			verify(service).getPostfachId(SAML_TOKEN);
		}

		@Test
		void shouldCallPostfachService() {
			findRueckfragen();

			verify(postfachNachrichtService).findRueckfragen(AntragraumService.BAYERN_ID_SERVICE_KONTO_TYPE, POSTFACH_ID);
		}

		@Test
		void shouldCallMapper() {
			findRueckfragen();

			verify(rueckfrageMapper).fromPostfachNachricht(postfachNachricht);
		}

		@Test
		void shouldCallEnrichByAnswers() {
			findRueckfragen();

			verify(service).enrichByAnswers(rueckfrage, SAML_TOKEN);
		}

		@Test
		void shouldCallVorgangService() {
			findRueckfragen();

			verify(vorgangService).getVorgang(RueckfrageHeadTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallEnrichByVorgangData() {
			findRueckfragen();

			verify(service).addVorgangData(rueckfrage, vorgang);
		}

		@Test
		void shouldAddAccessible() {
			findRueckfragen();

			verify(service).addAccessible(rueckfrage, SAML_TOKEN);
		}

		@Test
		void shouldReturnFullFilledValue() {
			var rueckfragen = findRueckfragen();

			assertThat(rueckfragen).containsExactly(rueckfrage);
		}

		private List<RueckfrageHead> findRueckfragen() {
			return service.findRueckfragen(SAML_TOKEN).toList();
		}
	}

	@DisplayName("Enrich by answers")
	@Nested
	class TestEnrichByAnswers {

		private static final String SAML_TOKEN = "TOKEN";
		private final RueckfrageHead rueckfrage = RueckfrageHeadTestFactory.create();
		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();
		private final Stream<PostfachNachricht> answers = Stream.of(postfachNachricht);

		@BeforeEach
		void mock() {
			doReturn(Optional.of(PostfachNachrichtTestFactory.SENT_AT)).when(service).getFirstSentAt(any());
			doReturn(answers).when(service).findAnswers(any(), any());
		}

		@Test
		void shouldCallFindAnswers() {
			enrichByAnswer();

			verify(service).findAnswers(SAML_TOKEN, RueckfrageHeadTestFactory.ID);
		}

		@Test
		void shouldSetAnsweredAt() {
			var enrichedRueckfrage = enrichByAnswer();

			assertThat(enrichedRueckfrage.getAnsweredAt()).isEqualTo(PostfachNachrichtTestFactory.SENT_AT);
		}

		@Test
		void shouldSetStatus() {
			var enrichedRueckfrage = enrichByAnswer();

			assertThat(enrichedRueckfrage.getStatus()).isEqualTo(RueckfrageStatus.ANSWERED);
		}

		@Test
		void shouldReturnFullFilledValue() {
			var enrichedRueckfrage = enrichByAnswer();

			assertThat(enrichedRueckfrage).usingRecursiveComparison().ignoringFields("answeredAt", "vorgangId", "vorgangName", "status")
					.isEqualTo(rueckfrage);
		}

		private RueckfrageHead enrichByAnswer() {
			return service.enrichByAnswers(rueckfrage, SAML_TOKEN);
		}
	}

	@DisplayName("Add accessible")
	@Nested
	class TestAddAccessible {

		private final RueckfrageHead rueckfrageHead = RueckfrageHeadTestFactory.createBuilder().accessible(false).build();

		@BeforeEach
		void mock() {
			doReturn(true).when(service).isAccessible(any(), any());
		}

		@Test
		void shouldCallIsAccessible() {
			service.addAccessible(rueckfrageHead, GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);

			verify(service).isAccessible(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN, GrpcServiceKontoTestFactory.TRUST_LEVEL);
		}

		@Test
		void shouldSetAccessible() {
			var rueckfrageWithAccessible = service.addAccessible(rueckfrageHead, GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);

			assertThat(rueckfrageWithAccessible.isAccessible()).isTrue();
		}
	}

	@DisplayName("Add vorgang data")
	@Nested
	class TestAddVorgangData {

		@Test
		void shouldSetVorgangName() {
			var rueckfragenWithVorgangData = addVorgangData();

			assertThat(rueckfragenWithVorgangData.getVorgangName()).isEqualTo(VorgangTestFactory.NAME);
		}

		@Test
		void shouldSetVorgangNummer() {
			var rueckfragenWithVorgangData = addVorgangData();

			assertThat(rueckfragenWithVorgangData.getVorgangNummer()).isEqualTo(VorgangTestFactory.NUMMER);
		}

		@Test
		void shouldSetTrustLevel() {
			var rueckfragenWithVorgangData = addVorgangData();

			assertThat(rueckfragenWithVorgangData.getTrustLevel()).isEqualTo(GrpcServiceKontoTestFactory.TRUST_LEVEL);
		}

		private RueckfrageHead addVorgangData() {
			return service.addVorgangData(RueckfrageHead.builder().build(), VorgangTestFactory.create());
		}
	}

	@DisplayName("Get first sentAt")
	@Nested
	class TestGetFirstSentAt {
		private static final ZonedDateTime earlyZonedDateTime = ZonedDateTime.parse("2020-01-01T10:00:10Z");
		private static final PostfachNachricht earlyPostfachNachricht = PostfachNachrichtTestFactory.createBuilder().sentAt(earlyZonedDateTime)
				.build();

		private static final ZonedDateTime lateZonedDateTime = ZonedDateTime.parse("2021-01-01T10:00:10Z");
		private static final PostfachNachricht latePostfachNachricht = PostfachNachrichtTestFactory.createBuilder().sentAt(lateZonedDateTime).build();

		@Test
		void shouldReturnSentAt() {
			var firstSentAt = service.getFirstSentAt(Stream.of(PostfachNachrichtTestFactory.create(), earlyPostfachNachricht, latePostfachNachricht));

			assertThat(firstSentAt).contains(earlyZonedDateTime);
		}
	}

	@DisplayName("Send rueckfragen answers")
	@Nested
	class TestSendRueckfrageAnswer {
		static final String SAML_TOKEN = "TOKEN";
		static final String RUECKFRAGE_ID = UUID.randomUUID().toString();
		static final PostfachNachricht NACHRICHT = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void mock() {
			doNothing().when(service).verifyToken(any());
			doNothing().when(service).verifyAccessToPostfach(any(), any());
		}

		@Test
		void shouldCallVerifyToken() {
			service.sendRueckfrageAnswer(SAML_TOKEN, RUECKFRAGE_ID, NACHRICHT);

			verify(service).verifyToken(SAML_TOKEN);
		}

		@Test
		void shouldCallPostfachService() {
			service.sendRueckfrageAnswer(SAML_TOKEN, RUECKFRAGE_ID, NACHRICHT);

			verify(postfachNachrichtService).persistAnswer(RUECKFRAGE_ID, NACHRICHT);
		}

		@Test
		void shouldVerifyAccessToPostfach() {
			service.sendRueckfrageAnswer(SAML_TOKEN, RUECKFRAGE_ID, NACHRICHT);

			verify(service).verifyAccessToPostfach(SAML_TOKEN, NACHRICHT);
		}
	}

	@Nested
	class TestFindAnswer {

		private static final String SAML_TOKEN = "TOKEN";

		@BeforeEach
		void init() {
			when(decrypter.decryptPostfachId(any())).thenReturn(GrpcFindRueckfrageRequestTestFactory.POSTFACH_ID);
		}

		@Test
		void shouldCallVerify() {
			service.findAnswers(SAML_TOKEN, PostfachNachrichtTestFactory.ID);

			verify(service).verifyToken(SAML_TOKEN);
		}

		@Test
		void shouldCallFindAnswers() {
			service.findAnswers(SAML_TOKEN, PostfachNachrichtTestFactory.ID);

			verify(postfachNachrichtService).findAnswers(AntragraumService.BAYERN_ID_SERVICE_KONTO_TYPE,
					GrpcFindRueckfrageRequestTestFactory.POSTFACH_ID,
					PostfachNachrichtTestFactory.ID);
		}
	}

	@DisplayName("Get postfachId")
	@Nested
	class TestGetPostfachId {

		@Mock
		private Response response;

		private static final String SAML_TOKEN = "TOKEN";

		@BeforeEach
		void init() {
			when(parser.parse(anyString())).thenReturn(response);
			when(decrypter.decryptPostfachId(any())).thenReturn(GrpcFindRueckfrageRequestTestFactory.POSTFACH_ID);
		}

		@Test
		void shouldCallParser() {
			service.getPostfachId(SAML_TOKEN);

			verify(parser).parse(SAML_TOKEN);
		}

		@Test
		void shouldCallDecrypter() {
			service.getPostfachId(SAML_TOKEN);

			verify(decrypter).decryptPostfachId(response);
		}
	}

	@DisplayName("Get rueckfrage")
	@Nested
	class TestGetRueckfrage {

		private final String samlToken = "DummySamlToken";

		private final Map<String, Object> nachrichtAsMap = PostfachNachrichtTestFactory.asMap();
		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void mock() {
			when(postfachNachrichtService.getById(any())).thenReturn(nachrichtAsMap);
			when(nachrichtMapper.fromMapToPostfachMail(any())).thenReturn(nachricht);

			doNothing().when(service).verifyToken(any());
			doNothing().when(service).verifyAccessToPostfach(any(), any());
		}

		@Test
		void shouldVerifySamlToken() {
			getRueckfrage();

			verify(service).verifyToken(samlToken);
		}

		@Test
		void shouldCallPersistPostfachNachrichtService() {
			getRueckfrage();

			verify(postfachNachrichtService).getById(RueckfrageHeadTestFactory.ID);
		}

		@Test
		void shouldCallNachrichtMapper() {
			getRueckfrage();

			verify(nachrichtMapper).fromMapToPostfachMail(nachrichtAsMap);
		}

		@Test
		void shouldReturnValue() {
			var rueckfrage = getRueckfrage();

			assertThat(rueckfrage).isEqualTo(nachricht);
		}

		@Test
		void shouldCallVerifyAccessToPostfach() {
			getRueckfrage();

			verify(service).verifyAccessToPostfach(samlToken, nachricht);
		}

		private PostfachNachricht getRueckfrage() {
			return service.getRueckfrage(samlToken, RueckfrageHeadTestFactory.ID);
		}
	}

	@Nested
	class TestVerifyPostfachId {

		private final String SAML_TOKEN = "TOKEN";
		private final PostfachNachricht POSTFACH_NACHRICHT = PostfachNachrichtTestFactory.create();
		private final Vorgang vorgang = VorgangTestFactory.create();

		@Test
		void shouldGetPostfachIdFromToken() {
			doReturn(GrpcPostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE).when(service).getPostfachId(any());

			service.verifyPostfachId(SAML_TOKEN, POSTFACH_NACHRICHT, vorgang);

			verify(service).getPostfachId(SAML_TOKEN);
		}

		@Test
		void shouldThrowExceptionOnMismatchPostfachId() {
			doReturn("not-matching-postfach-id").when(service).getPostfachId(any());

			assertThatThrownBy(() -> service.verifyPostfachId(SAML_TOKEN, POSTFACH_NACHRICHT, vorgang))
					.isInstanceOf(NotFoundException.class)
					.hasMessageContaining("PostfachNachricht")
					.hasMessageContaining(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldNotThrowExceptionOnMatchingPostfachId() {
			doReturn(GrpcPostfachAddressTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE).when(service).getPostfachId(any());

			assertDoesNotThrow(() -> service.verifyPostfachId(SAML_TOKEN, POSTFACH_NACHRICHT, vorgang));
		}
	}

	@DisplayName("Is accessible")
	@Nested
	class TestIsAccessible {

		@Test
		void shouldCallGetTrustLevel() {
			doReturn(TrustLevel.LEVEL_1).when(service).getTrustLevel(any());

			service.isAccessible(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN, TrustLevel.LEVEL_1.getValue());

			verify(service).getTrustLevel(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);
		}

		@Test
		void shouldReturnTrueIfTrustLevelMatches() {
			doReturn(TrustLevel.LEVEL_1).when(service).getTrustLevel(any());

			var trustLevel = service.isAccessible(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN, TrustLevel.LEVEL_1.getValue());

			assertThat(trustLevel).isTrue();
		}

		@Test
		void shouldAllowAccessOnHigherTrustLevel() {
			doReturn(TrustLevel.LEVEL_2).when(service).getTrustLevel(any());

			var trustLevel = service.isAccessible(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN, TrustLevel.LEVEL_1.getValue());

			assertThat(trustLevel).isTrue();
		}
	}

	@DisplayName("Get trustLevel")
	@Nested
	class TestGetTrustLevel {

		@Mock
		private Response response;

		@BeforeEach
		void mock() {
			doReturn(response).when(service).parseSamlToken(any());

			when(decrypter.decryptTrustLevel(any())).thenReturn(TrustLevel.LEVEL_1.getValue());
		}

		@Test
		void shouldCallParseSamlToken() {
			service.getTrustLevel(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);

			verify(service).parseSamlToken(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);
		}

		@Test
		void shouldCallDecrypter() {
			service.getTrustLevel(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);

			verify(decrypter).decryptTrustLevel(response);
		}

		@Test
		void shouldReturnValue() {
			var trustLevel = service.getTrustLevel(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);

			assertThat(trustLevel).isEqualTo(TrustLevel.LEVEL_1);
		}
	}

	@DisplayName("Parse saml token")
	@Nested
	class TestParseSamlToken {

		@Mock
		private Response response;

		@BeforeEach
		void mock() {
			when(parser.parse(any())).thenReturn(response);
		}

		@Test
		void shouldCallParser() {
			service.parseSamlToken(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);

			verify(parser).parse(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);
		}

		@Test
		void shouldReturnValue() {
			var parseResponse = service.parseSamlToken(GrpcGetRueckfrageRequestTestFactory.SAML_TOKEN);

			assertThat(parseResponse).isEqualTo(response);
		}
	}

	@Nested
	class TestGetAttachmentContent {

		@Mock
		private OutputStream outputStream;

		private final AttachmentFileRequest request = AttachmentFileRequestTestFactory.create();

		@BeforeEach
		void mock() {
			doNothing().when(service).verifyAccessToFile(request);
		}

		@Test
		void shouldCallVerifyAccessToFile() {
			callGetAttachmentContent();

			verify(service).verifyAccessToFile(request);
		}

		@Test
		void shouldCallOzgCloudFileServiceToWriteData() {
			callGetAttachmentContent();

			verify(ozgCloudFileService).writeFileDataToStream(OzgCloudFileId.from(AttachmentFileRequestTestFactory.FILE_ID), outputStream);
		}

		@SneakyThrows
		private void callGetAttachmentContent() {
			service.getAttachmentContent(request, outputStream);
		}
	}

	@Nested
	class TestGetAttachmentMetadata {

		private final AttachmentFileRequest request = AttachmentFileRequestTestFactory.create();

		@BeforeEach
		void mock() {
			doNothing().when(service).verifyAccessToFile(request);
		}

		@Test
		void shouldCallVerifyAccessToFile() {
			callGetAttachmentMetadata();

			verify(service).verifyAccessToFile(request);
		}

		@Test
		void shouldCallOzgCloudFileServiceToGetFile() {
			callGetAttachmentMetadata();

			verify(ozgCloudFileService).getFile(OzgCloudFileId.from(AttachmentFileRequestTestFactory.FILE_ID));
		}

		@Test
		void shouldReturnOzgCloudFile() {
			var ozgCloudFile = OzgCloudFileTestFactory.create();
			when(ozgCloudFileService.getFile(OzgCloudFileId.from(AttachmentFileRequestTestFactory.FILE_ID))).thenReturn(ozgCloudFile);

			var result = callGetAttachmentMetadata();

			assertThat(result).isEqualTo(ozgCloudFile);
		}

		private OzgCloudFile callGetAttachmentMetadata() {
			return service.getAttachmentMetadata(request);
		}
	}

	@Nested
	class TestVerifyAccessToFile {

		private final AttachmentFileRequest request = AttachmentFileRequestTestFactory.create();
		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void mock() {
			when(attachedItemService.getPostfachNachricht(AttachmentFileRequestTestFactory.NACHRICHT_ID)).thenReturn(nachricht);
			doNothing().when(service).verifyAccessToPostfach(AttachmentFileRequestTestFactory.TOKEN, nachricht);
			doNothing().when(service).verifyFileId(AttachmentFileRequestTestFactory.FILE_ID, nachricht);
		}

		@Test
		void shouldVerifySamlToken() {
			callVerifyAccessToFile();

			verify(service).verifyToken(AttachmentFileRequestTestFactory.TOKEN);
		}

		@Test
		void shouldCallAttachedItemServiceToGetPostfachNachricht() {
			callVerifyAccessToFile();

			verify(attachedItemService).getPostfachNachricht(AttachmentFileRequestTestFactory.NACHRICHT_ID);
		}

		@Test
		void shouldCallVerifyAccessToPostfach() {
			callVerifyAccessToFile();

			verify(service).verifyAccessToPostfach(AttachmentFileRequestTestFactory.TOKEN, nachricht);
		}

		@Test
		void shouldVerifyFileId() {
			callVerifyAccessToFile();

			verify(service).verifyFileId(AttachmentFileRequestTestFactory.FILE_ID, nachricht);
		}

		private void callVerifyAccessToFile() {
			service.verifyAccessToFile(request);
		}

	}

	@Nested
	class TestVerifyAccessToPostfach {

		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.create();
		private final Vorgang vorgang = VorgangTestFactory.create();

		@BeforeEach
		void mock() {
			when(vorgangService.getVorgang(MessageTestFactory.VORGANG_ID)).thenReturn(vorgang);
			doNothing().when(service).verifyPostfachId(AttachmentFileRequestTestFactory.TOKEN, nachricht, vorgang);
			doNothing().when(service).verifyTrustLevel(AttachmentFileRequestTestFactory.TOKEN, nachricht, vorgang);
		}

		@Test
		void shouldCallVorgangService() {
			verifyAccessToPostfach();

			verify(vorgangService).getVorgang(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallVerifyPostfachId() {
			verifyAccessToPostfach();

			verify(service).verifyPostfachId(AttachmentFileRequestTestFactory.TOKEN, nachricht, vorgang);
		}

		@Test
		void shouldVerifyTrustLevel() {
			verifyAccessToPostfach();

			verify(service).verifyTrustLevel(AttachmentFileRequestTestFactory.TOKEN, nachricht, vorgang);
		}

		private void verifyAccessToPostfach() {
			service.verifyAccessToPostfach(AttachmentFileRequestTestFactory.TOKEN, nachricht);
		}
	}

	@Nested
	class TestVerifyTrustLevel {

		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.create();
		private final Vorgang vorgang = VorgangTestFactory.create();

		@Test
		void shouldCallIsAccessible() {
			doReturn(true).when(service).isAccessible(AttachmentFileRequestTestFactory.TOKEN, GrpcServiceKontoTestFactory.TRUST_LEVEL);

			callVerifyTrustLevel();

			verify(service).isAccessible(AttachmentFileRequestTestFactory.TOKEN, GrpcServiceKontoTestFactory.TRUST_LEVEL);
		}

		@Test
		void shouldThrowExceptionOnNotAccessible() {
			doReturn(false).when(service).isAccessible(AttachmentFileRequestTestFactory.TOKEN, GrpcServiceKontoTestFactory.TRUST_LEVEL);

			assertThatThrownBy(() -> callVerifyTrustLevel())
					.isInstanceOf(NotFoundException.class)
					.hasMessageContaining("PostfachNachricht")
					.hasMessageContaining(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldNotThrowExceptionOnAccessible() {
			doReturn(true).when(service).isAccessible(AttachmentFileRequestTestFactory.TOKEN, GrpcServiceKontoTestFactory.TRUST_LEVEL);

			assertDoesNotThrow(() -> callVerifyTrustLevel());
		}

		private void callVerifyTrustLevel() {
			service.verifyTrustLevel(AttachmentFileRequestTestFactory.TOKEN, nachricht, vorgang);
		}
	}

	@Nested
	class TestVerifyFileId {
		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.create();

		@Test
		void shouldThrowExceptionOnNotMatchingFileId() {
			assertThatThrownBy(() -> service.verifyFileId(PostfachNachrichtTestFactory.ATTACHMENT_FILE_ID + "wrong", nachricht))
					.isInstanceOf(NotFoundException.class)
					.hasMessageContaining("PostfachNachricht")
					.hasMessageContaining(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldNotThrowExceptionOnMatchingFileId() {
			assertDoesNotThrow(() -> service.verifyFileId(PostfachNachrichtTestFactory.ATTACHMENT_FILE_ID, nachricht));
		}
	}
}