/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.antragraum;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.NoSuchElementException;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.opensaml.saml.saml2.core.AttributeStatement;
import org.opensaml.saml.saml2.core.Response;
import org.springframework.core.io.InputStreamResource;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.common.test.TestUtils;

@Disabled("Passende certificat und key file muessen local vorhanden sein")
class Saml2DecrypterTest {
	private Response samlResponse;

	@Mock
	private AntragraumProperties properties;

	private BayernIdSamlConfiguration configuration;

	private Saml2Decrypter decrypter;

	@BeforeEach
	void setup() throws Exception {
		when(properties.getDecryptionCertificate()).thenReturn(new InputStreamResource(TestUtils.loadFile("bayernid-test-enc.crt")));
		when(properties.getDecryptionPrivateKey()).thenReturn(new InputStreamResource(TestUtils.loadFile("bayernid-test-enc.key")));

		configuration = new BayernIdSamlConfiguration();
		ReflectionTestUtils.setField(configuration, "antragraumProperties", properties);
		configuration.initOpenSAML();

		var samlResponseString = TestUtils.loadTextFile("SamlResponse.xml");
		var parser = new Saml2Parser(configuration);
		samlResponse = parser.parse(samlResponseString);

		decrypter = Mockito.spy(new Saml2Decrypter(configuration));
		decrypter.init();
	}

	@Test
	void shouldDecrypt() {
		assertThat(samlResponse.getAssertions()).isEmpty();

		decrypter.decryptResponseElements(samlResponse);

		assertThat(samlResponse.getAssertions()).isNotEmpty();
	}

	@Test
	void shouldHaveSubject() {
		decrypter.decryptResponseElements(samlResponse);
		var samlAssertion = samlResponse.getAssertions().get(0);

		assertThat(samlAssertion.getSubject()).isNotNull();
	}

	@Test
	void shouldHaveAuthnStatements() {
		decrypter.decryptResponseElements(samlResponse);
		var samlAssertion = samlResponse.getAssertions().get(0);
		var authnStatements = samlAssertion.getAuthnStatements();

		assertThat(authnStatements).isNotNull();
	}

	@Test
	void shouldHaveStatements() {
		decrypter.decryptResponseElements(samlResponse);
		var samlAssertion = samlResponse.getAssertions().get(0);
		var statements = samlAssertion.getStatements();

		assertThat(statements).isNotNull();
	}

	@Test
	void shouldHaveAttributes() {
		decrypter.decryptResponseElements(samlResponse);
		var samlAssertion = samlResponse.getAssertions().get(0);
		var statements = (AttributeStatement) samlAssertion.getStatements().get(1);
		var attributes = statements.getAttributes();
		assertThat(attributes).hasSize(7);
	}

	@DisplayName("Decrypt postfachId")
	@Nested
	class TestDecryptPostfachId {

		@Test
		void shouldReturnPostfachIdIfExists() {
			var postfachId = decrypter.decryptPostfachId(samlResponse);

			assertThat(postfachId).isEqualTo("28721c6f-b78f-4d5c-a048-19fd2fc429d2");
		}

		@Test
		void shouldThrowExceptionIfMissing() {
			doReturn(false).when(decrypter).isPostfachId(any());

			assertThatThrownBy(() -> decrypter.decryptPostfachId(samlResponse)).isInstanceOf(NoSuchElementException.class);
		}
	}

	@DisplayName("Decrypt trustLevel")
	@Nested
	class TestDecryptTrustLevel {

		@Test
		void shouldReturnTrustLevelIfExists() {
			var trustLevel = decrypter.decryptTrustLevel(samlResponse);

			assertThat(trustLevel).isEqualTo("STORK-QAA-Level-1");
		}

		@Test
		void shouldThrowExceptionIfMissing() {
			doReturn(false).when(decrypter).isTrustLevel(any());

			assertThatThrownBy(() -> decrypter.decryptTrustLevel(samlResponse)).isInstanceOf(NoSuchElementException.class);
		}
	}
}