/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.nachrichten.common.grpc;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.UUID;

import org.apache.logging.log4j.ThreadContext;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Mockito;

import de.ozgcloud.common.grpc.GrpcUtil;
import io.grpc.ClientCall;
import io.grpc.Metadata;

class NachrichtenCallContextAttachingInterceptorTest {

	@InjectMocks
	private NachrichtenCallContextAttachingInterceptor interceptor;

	@Mock
	private ClientCall<RequestTest, ResponseTest> delegateCall;
	@Mock
	private ClientCall.Listener<ResponseTest> responseTestListener;
	@Mock
	private Metadata requestMetadata;

	private NachrichtenCallContextAttachingInterceptor.CallContextAttachingClientCall<RequestTest, ResponseTest> clientCall;

	@BeforeEach
	void init() {
		clientCall = spy(interceptor.new CallContextAttachingClientCall<>(delegateCall));
	}

	@Nested
	class TestStart {

		@Mock
		private Metadata customMetadata;

		@BeforeEach
		void init() {
			doReturn(customMetadata).when(clientCall).buildCallContextMetadata();
		}

		@Test
		void shouldCallBuildCallContextMetadata() {
			clientCall.start(responseTestListener, requestMetadata);

			verify(clientCall).buildCallContextMetadata();
		}

		@Test
		void shouldMergeMetadata() {
			clientCall.start(responseTestListener, requestMetadata);

			verify(requestMetadata).merge(customMetadata);
		}
	}

	@Nested
	class TestBuildCallContextMetadata {

		@Test
		void shouldSetUserId() {
			var result = clientCall.buildCallContextMetadata();

			assertThat(result.get(GrpcUtil.HEADER_KEY_USER_ID))
					.isEqualTo(NachrichtenCallContextAttachingInterceptor.NACHRICHTEN_MANAGER_SENDER_USER_ID.getBytes());
		}

		@Test
		void shouldSetClientName() {
			var result = clientCall.buildCallContextMetadata();

			assertThat(result.get(GrpcUtil.HEADER_KEY_CLIENT_NAME))
					.isEqualTo(NachrichtenCallContextAttachingInterceptor.NACHRICHTEN_MANAGER_CLIENT_NAME.getBytes());
		}

		@Test
		void shouldCallGetrequestId() {
			clientCall.buildCallContextMetadata();

			verify(clientCall).getRequestId();
		}

		@Test
		void shouldSetRequestId() {
			var requestId = UUID.randomUUID().toString().getBytes();
			doReturn(requestId).when(clientCall).getRequestId();

			var result = clientCall.buildCallContextMetadata();

			assertThat(result.get(GrpcUtil.HEADER_KEY_REQUEST_ID)).isEqualTo(requestId);
		}
	}

	@Nested
	class TestGetRequestId {

		private MockedStatic<ThreadContext> threadContextMocked;

		@BeforeEach
		void init() {
			threadContextMocked = Mockito.mockStatic(ThreadContext.class);
		}

		@AfterEach
		void close() {
			threadContextMocked.close();
		}

		@Test
		void shouldCallGetContext() {
			clientCall.getRequestId();

			threadContextMocked.verify(ThreadContext::getContext);
		}

		@Test
		void shouldReturnRequestIdFromThreadContext() {
			var requestId = UUID.randomUUID().toString();

			threadContextMocked.when(ThreadContext::getContext)
					.thenReturn(Map.of(NachrichtenCallContextAttachingInterceptor.REQUEST_ID_KEY, requestId));

			var result = clientCall.getRequestId();

			assertThat(result).isEqualTo(requestId.getBytes());
		}

		@Test
		void shouldReturnNewRequestId() {
			var result = clientCall.getRequestId();

			assertThat(result).isNotNull();
		}
	}

	private record RequestTest() {
	}

	private record ResponseTest() {
	}
}