/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.file;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PipedInputStream;
import java.net.FileNameMap;
import java.net.URLConnection;
import java.util.Base64;
import java.util.Optional;
import java.util.UUID;
import java.util.function.Consumer;

import jakarta.activation.MimetypesFileTypeMap;

import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;
import org.springframework.core.task.TaskExecutor;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.apilib.file.OzgCloudFile;
import de.ozgcloud.apilib.file.OzgCloudFileId;
import de.ozgcloud.apilib.file.OzgCloudFileService;
import de.ozgcloud.apilib.file.OzgCloudUploadFile;
import de.ozgcloud.apilib.vorgang.OzgCloudFileIdMapper;
import de.ozgcloud.common.test.TestUtils;
import de.ozgcloud.nachrichten.postfach.FileId;
import lombok.SneakyThrows;

class AttachmentFileServiceTest {

	private static final String FILE_ID = UUID.randomUUID().toString();
	private static final byte[] FILE_CONTENT = LoremIpsum.getInstance().getWords(1).getBytes();

	@Spy
	@InjectMocks
	private AttachmentFileService service;

	@Mock
	private OzgCloudFileService ozgCloudFileService;
	@Mock
	private OzgCloudFileIdMapper fileIdMapper;
	@Mock
	private AttachmentFileMapper attachmentFileMapper;
	@Mock
	private TaskExecutor taskExecutor;
	@Mock
	private MimetypesFileTypeMap mimetypesFileTypeMap;

	@Nested
	class TestGetFile {

		@Mock
		private OzgCloudFile ozgCloudFile;
		@Mock
		private AttachmentFile attachmentFile;

		@BeforeEach
		void init() {
			when(fileIdMapper.toFileId(any())).thenReturn(OzgCloudFileId.from(FILE_ID));
			when(attachmentFileMapper.fromOzgCloudFile(any())).thenReturn(attachmentFile);
			when(ozgCloudFileService.getFile(any())).thenReturn(ozgCloudFile);
		}

		@Test
		void shouldCallToFileId() {
			getFile();

			verify(fileIdMapper).toFileId(FILE_ID);
		}

		@Test
		void shouldCallGetFile() {
			getFile();

			verify(ozgCloudFileService).getFile(OzgCloudFileId.from(FILE_ID));
		}

		@Test
		void shouldCallFromOzgCloudFile() {
			getFile();

			verify(attachmentFileMapper).fromOzgCloudFile(ozgCloudFile);
		}

		@Test
		void shouldReturnAttachmentFile() {
			var result = getFile();

			assertThat(result).isEqualTo(attachmentFile);
		}

		private AttachmentFile getFile() {
			return service.getFile(FileId.from(FILE_ID));
		}
	}

	@Nested
	class TestGetFileContent {

		@Mock
		private PipedInputStream inputStream;
		@Mock
		private OutputStream outputStream;

		@BeforeEach
		void init() {
			doReturn(inputStream).when(service).createInputStream();
			doReturn(outputStream).when(service).connectToOutputStream(any());
			doNothing().when(service).writeFileContent(any(), any());
		}

		@Test
		void shouldCallCreateInputStream() {
			getFileContent();

			verify(service).createInputStream();
		}

		@Test
		void shouldCallConnectToOutputStream() {
			getFileContent();

			verify(service).connectToOutputStream(inputStream);
		}

		@Test
		void shoulcCallWriteFileContent() {
			getFileContent();

			verify(service).writeFileContent(FILE_ID, outputStream);
		}

		@Test
		void shouldReturnContentStream() {
			var result = getFileContent();

			assertThat(result).isEqualTo(inputStream);
		}

		private InputStream getFileContent() {
			return service.getFileContent(FileId.from(FILE_ID));
		}
	}

	@Nested
	class TestConnectToOutputStream {

		@SneakyThrows
		@Test
		void shouldReturnOutputStream() {
			try (var inputStream = new PipedInputStream()) {

				var result = service.connectToOutputStream(inputStream);
				result.write(FILE_CONTENT);
				result.close();

				assertThat(inputStream).hasBinaryContent(FILE_CONTENT);
			}
		}
	}

	@Nested
	class TestWriteFileContent {

		private static final OzgCloudFileId OZG_CLOUD_FILE_ID = OzgCloudFileId.from(FILE_ID);

		@Mock
		private OutputStream outputStream;

		@Captor
		private ArgumentCaptor<Runnable> runnableCaptor;

		@BeforeEach
		void init() {
			when(fileIdMapper.toFileId(any())).thenReturn(OZG_CLOUD_FILE_ID);
		}

		@Test
		void shouldCallToFileId() {
			writeFileContent();

			verify(fileIdMapper).toFileId(FILE_ID);
		}

		@Test
		void shouldCallWriteFileDataToStream() {
			writeFileContent();

			verify(ozgCloudFileService).writeFileDataToStream(OZG_CLOUD_FILE_ID, outputStream);
		}

		@Test
		void shouldCallCloseOutputStream() {
			try (var ioUtilsMock = mockStatic(IOUtils.class)) {
				writeFileContent();

				ioUtilsMock.verify(() -> IOUtils.closeQuietly(eq(outputStream), any(Consumer.class)));
			}
		}

		private void writeFileContent() {
			service.writeFileContent(FILE_ID, outputStream);
			verify(taskExecutor).execute(runnableCaptor.capture());
			runnableCaptor.getValue().run();
		}

	}

	@Nested
	class TestUploadAttachmentFile {

		private static final AttachmentFile ATTACHMENT_FILE = AttachmentFileTestFactory.create();
		private static final String FILE_CONTENT_STR = new String(FILE_CONTENT);

		@Mock
		private AttachmentFile enhancedAttachmentFile;
		@Mock
		private OzgCloudUploadFile ozgCloudUploadFile;
		@Mock
		private InputStream contentStream;
		@Mock
		private Base64.Decoder decoder;

		private MockedStatic<Base64> base64Mock;

		@BeforeEach
		void init() {
			when(ozgCloudFileService.uploadFile(any(), any())).thenReturn(OzgCloudFileId.from(FILE_ID));
			when(attachmentFileMapper.toOzgCloudUploadFile(any())).thenReturn(ozgCloudUploadFile);
			doReturn(enhancedAttachmentFile).when(service).addContentTypeIfMissing(any(), any());
			doReturn(contentStream).when(service).toInputStream(any());
			base64Mock = mockStatic(Base64.class);
			base64Mock.when(Base64::getDecoder).thenReturn(decoder);
			when(decoder.decode(anyString())).thenReturn(FILE_CONTENT);
		}

		@AfterEach
		void close() {
			base64Mock.close();
		}

		@Test
		void shouldCallGetDecoder() {
			uploadAttachmentFile();

			base64Mock.verify(Base64::getDecoder);
		}

		@Test
		void shouldCallDecode() {
			uploadAttachmentFile();

			verify(decoder).decode(FILE_CONTENT_STR);
		}

		@Test
		void shouldCallAddContentTypeIfMissing() {
			uploadAttachmentFile();

			verify(service).addContentTypeIfMissing(ATTACHMENT_FILE, FILE_CONTENT);
		}

		@Test
		void shouldCallToOzgCloudUploadFile() {
			uploadAttachmentFile();

			verify(attachmentFileMapper).toOzgCloudUploadFile(enhancedAttachmentFile);
		}

		@Test
		void shouldCallToDecodedInputStream() {
			uploadAttachmentFile();

			verify(service).toInputStream(FILE_CONTENT);
		}

		@Test
		void shouldCallUploadFile() {
			uploadAttachmentFile();

			verify(ozgCloudFileService).uploadFile(ozgCloudUploadFile, contentStream);
		}

		@Test
		void shouldReturnFileId() {
			var result = uploadAttachmentFile();

			assertThat(result).isEqualTo(FILE_ID);
		}

		private String uploadAttachmentFile() {
			return service.uploadAttachmentFile(ATTACHMENT_FILE, FILE_CONTENT_STR);
		}
	}

	@Nested
	class TestAddContentTypeIfMissing {

		private AttachmentFile attachmentFile;

		@Nested
		class TestNoContentType {

			@BeforeEach
			void init() {
				attachmentFile = AttachmentFileTestFactory.createBuilder().contentType(null).build();
				doReturn(AttachmentFileTestFactory.CONTENT_TYPE).when(service).getContentType(any(), any());
			}

			@Test
			void shouldCallGetContentType() {
				addContentTypeIfMissing();

				verify(service).getContentType(attachmentFile, FILE_CONTENT);
			}

			@Test
			void shouldReturnAttachmentFile() {
				var result = addContentTypeIfMissing();

				assertThat(result).usingRecursiveComparison().isEqualTo(AttachmentFileTestFactory.create());
			}
		}

		@Test
		void shouldReturnAttachmentFile() {
			attachmentFile = AttachmentFileTestFactory.create();

			var result = addContentTypeIfMissing();

			assertThat(result).isSameAs(attachmentFile);
		}

		private AttachmentFile addContentTypeIfMissing() {
			return service.addContentTypeIfMissing(attachmentFile, FILE_CONTENT);
		}
	}

	@Nested
	class TestGetContentType {

		private static final AttachmentFile ATTACHMENT_FILE = AttachmentFileTestFactory.create();

		@Nested
		class TestGetByFileName {

			@BeforeEach
			void init() {
				doReturn(Optional.of(AttachmentFileTestFactory.CONTENT_TYPE)).when(service).getTypeByFileName(any());
			}

			@Test
			void shouldCallGetTypeByFileName() {
				getContentType();

				verify(service).getTypeByFileName(ATTACHMENT_FILE);
			}

			@Test
			void shouldReturnContentType() {
				var result = getContentType();

				assertThat(result).isEqualTo(AttachmentFileTestFactory.CONTENT_TYPE);
			}
		}

		@Nested
		class TestTypeByContent {

			@BeforeEach
			void init() {
				doReturn(Optional.empty()).when(service).getTypeByFileName(any());
				doReturn(Optional.of(AttachmentFileTestFactory.CONTENT_TYPE)).when(service).getTypeByContent(any());
			}

			@Test
			void shouldCallGetTypeByContent() {
				getContentType();

				verify(service).getTypeByContent(FILE_CONTENT);
			}

			@Test
			void shouldReturnContentType() {
				var result = getContentType();

				assertThat(result).isEqualTo(AttachmentFileTestFactory.CONTENT_TYPE);
			}

		}

		@Nested
		class TestByMimeTypes {

			@BeforeEach
			void init() {
				doReturn(Optional.empty()).when(service).getTypeByFileName(any());
				doReturn(Optional.empty()).when(service).getTypeByContent(any());
				doReturn(AttachmentFileTestFactory.CONTENT_TYPE).when(service).getByMimeTypes(any());
			}

			@Test
			void shouldCallGetByMimeTypes() {
				getContentType();

				verify(service).getByMimeTypes(ATTACHMENT_FILE);
			}

			@Test
			void shouldReturnContentType() {
				var result = getContentType();

				assertThat(result).isEqualTo(AttachmentFileTestFactory.CONTENT_TYPE);
			}
		}

		private String getContentType() {
			return service.getContentType(ATTACHMENT_FILE, FILE_CONTENT);
		}
	}

	@Nested
	class TestGetTypeByFileName {

		private static final AttachmentFile ATTACHMENT_FILE = AttachmentFileTestFactory.create();

		@Mock
		private FileNameMap fileNameMap;
		private MockedStatic<URLConnection> urlConnectionMock;

		@BeforeEach
		void init() {
			urlConnectionMock = mockStatic(URLConnection.class);
			urlConnectionMock.when(URLConnection::getFileNameMap).thenReturn(fileNameMap);
		}

		@AfterEach
		void close() {
			urlConnectionMock.close();
		}

		@Test
		void shouldCallGetContentTypeFor() {
			getTypeByFileName();

			verify(fileNameMap).getContentTypeFor(AttachmentFileTestFactory.NAME);
		}

		@Test
		void shouldReturnContentType() {
			when(fileNameMap.getContentTypeFor(any())).thenReturn(AttachmentFileTestFactory.CONTENT_TYPE);

			var result = getTypeByFileName();

			assertThat(result).contains(AttachmentFileTestFactory.CONTENT_TYPE);
		}

		@Test
		void shouldReturnEmpty() {
			var result = getTypeByFileName();

			assertThat(result).isEmpty();
		}

		private Optional<String> getTypeByFileName() {
			return service.getTypeByFileName(ATTACHMENT_FILE);
		}
	}

	@Nested
	class TestGetTypeByContent {

		@SneakyThrows
		@Test
		void shouldReturnContentType() {
			var content = TestUtils.loadFile("BspQuittung.xml").readAllBytes();

			var result = service.getTypeByContent(content);

			assertThat(result).contains("application/xml");
		}

		@Test
		void shouldReturnEmpty() {
			var result = service.getTypeByContent(FILE_CONTENT);

			assertThat(result).isEmpty();
		}

		@Test
		void shouldNotThrowException() {
			try (var urlConnectionMock = mockStatic(URLConnection.class)) {
				urlConnectionMock.when(() -> URLConnection.guessContentTypeFromStream(any())).thenThrow(IOException.class);

				var result = service.getTypeByContent(FILE_CONTENT);

				assertThat(result).isEmpty();
			}

		}

		@Nested
		class TestGetByMimeTypes {

			@Test
			void shouldCallGetContentType() {
				getByMimeTypes();

				verify(mimetypesFileTypeMap).getContentType(AttachmentFileTestFactory.NAME);
			}

			@Test
			void shouldReturnContentType() {
				when(mimetypesFileTypeMap.getContentType(anyString())).thenReturn(AttachmentFileTestFactory.CONTENT_TYPE);

				var result = getByMimeTypes();

				assertThat(result).isEqualTo(AttachmentFileTestFactory.CONTENT_TYPE);
			}

			private String getByMimeTypes() {
				return service.getByMimeTypes(AttachmentFileTestFactory.create());
			}
		}
	}

	@Nested
	class TestToInputStream {

		@Test
		void shouldReturnInputStream() {
			var result = service.toInputStream(FILE_CONTENT);

			assertThat(result).hasBinaryContent(FILE_CONTENT);
		}
	}
}