/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.nachrichten.common.grpc.NachrichtenCallContextAttachingInterceptor;
import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;
import de.ozgcloud.vorgang.common.GrpcObject;
import de.ozgcloud.vorgang.common.GrpcProperty;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemResponse;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItem;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemResponse;
import de.ozgcloud.vorgang.vorgangAttachedItem.VorgangAttachedItemServiceGrpc.VorgangAttachedItemServiceBlockingStub;

class AttachedItemRemoteServiceTest {

	@Spy
	@InjectMocks
	private AttachedItemRemoteService service;
	@Mock
	private VorgangAttachedItemServiceBlockingStub vorgangAttachedItemServiceBlockingStub;
	@Mock
	private PostfachNachrichtMapper postfachNachrichtMapper;

	@Nested
	class TestFindReplyAllowedPostfachNachrichts {

		private static final GrpcVorgangAttachedItem GRPC_VORGANG_ATTACHED_ITEM = GrpcVorgangAttachedItemTestFactory.create();

		@Mock
		private GrpcFindVorgangAttachedItemRequest grpcFindVorgangAttachedItemRequest;
		@Mock
		private GrpcFindVorgangAttachedItemResponse grpcFindVorgangAttachedItemResponse;
		@Mock
		private PostfachNachricht postfachNachricht;

		@BeforeEach
		void init() {
			when(grpcFindVorgangAttachedItemResponse.getVorgangAttachedItemsList()).thenReturn(Collections.singletonList(GRPC_VORGANG_ATTACHED_ITEM));
			when(vorgangAttachedItemServiceBlockingStub.find(any())).thenReturn(grpcFindVorgangAttachedItemResponse);
			doReturn(vorgangAttachedItemServiceBlockingStub).when(service).getVorgangAttachedItemServiceStub();
		}

		@Test
		void shouldCallGetVorgangAttachedItemServiceStub() {
			service.findReplyAllowedPostfachNachrichts(MessageTestFactory.VORGANG_ID);

			verify(service).getVorgangAttachedItemServiceStub();
		}

		@Test
		void shouldCallBuildFindRequest() {
			service.findReplyAllowedPostfachNachrichts(MessageTestFactory.VORGANG_ID);

			verify(service).buildFindRequest(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallFind() {
			doReturn(grpcFindVorgangAttachedItemRequest).when(service).buildFindRequest(anyString());

			service.findReplyAllowedPostfachNachrichts(MessageTestFactory.VORGANG_ID);

			verify(vorgangAttachedItemServiceBlockingStub).find(grpcFindVorgangAttachedItemRequest);
		}

		@Test
		void shouldCallReplyAllowed() {
			service.findReplyAllowedPostfachNachrichts(MessageTestFactory.VORGANG_ID).toList();

			verify(service).replyAllowed(GRPC_VORGANG_ATTACHED_ITEM);
		}

		@Test
		void shouldCallPostfachNachrichtMapper() {
			doReturn(true).when(service).replyAllowed(any());

			service.findReplyAllowedPostfachNachrichts(MessageTestFactory.VORGANG_ID).toList();

			verify(postfachNachrichtMapper).fromAttachedItem(GRPC_VORGANG_ATTACHED_ITEM);
		}

		@Test
		void shouldNotCallPostfachNachrichtMapper() {
			doReturn(false).when(service).replyAllowed(any());

			service.findReplyAllowedPostfachNachrichts(MessageTestFactory.VORGANG_ID).toList();

			verify(postfachNachrichtMapper, never()).fromAttachedItem(GRPC_VORGANG_ATTACHED_ITEM);
		}

		@Test
		void shouldReturnResult() {
			doReturn(true).when(service).replyAllowed(any());
			doReturn(postfachNachricht).when(postfachNachrichtMapper).fromAttachedItem(any());

			var result = service.findReplyAllowedPostfachNachrichts(MessageTestFactory.VORGANG_ID);

			assertThat(result).containsExactly(postfachNachricht);
		}
	}

	@Nested
	class TestReplyAllowed {

		@DisplayName("should return true")
		@ParameterizedTest(name = "when replyOption is {0}")
		@EnumSource(value = PostfachNachricht.ReplyOption.class, names = { "FORBIDDEN" }, mode = EnumSource.Mode.EXCLUDE)
		void shouldReturnTrue(PostfachNachricht.ReplyOption replyOption) {
			var grpcVorgangAttachedItem = GrpcVorgangAttachedItemTestFactory.createBuilder().setItem(GrpcObject.newBuilder()
					.addProperty(GrpcProperty.newBuilder().setName("replyOption").addValue(replyOption.name()))).build();

			var result = service.replyAllowed(grpcVorgangAttachedItem);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalse() {
			var grpcVorgangAttachedItem = GrpcVorgangAttachedItemTestFactory.createBuilder().setItem(GrpcObject.newBuilder()
					.addProperty(GrpcProperty.newBuilder().setName("replyOption").addValue("FORBIDDEN"))).build();

			var result = service.replyAllowed(grpcVorgangAttachedItem);

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestBuildFindRequest {

		@Test
		void shouldSetVorgangId() {
			var result = service.buildFindRequest(MessageTestFactory.VORGANG_ID);

			assertThat(result.getVorgangId()).isEqualTo(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetClient() {
			var result = service.buildFindRequest(MessageTestFactory.VORGANG_ID);

			assertThat(result.getClient()).isEqualTo(AttachedItemRemoteService.CLIENT);
		}

		@Test
		void shouldSetItemName() {
			var result = service.buildFindRequest(MessageTestFactory.VORGANG_ID);

			assertThat(result.getItemName()).isEqualTo(AttachedItemRemoteService.ITEM_NAME);
		}
	}

	@Nested
	class TestGetVorgangAttachedItemServiceStub {

		@Mock
		private VorgangAttachedItemServiceBlockingStub stubWithInterceptor;

		@Test
		void shouldAddInterceptor() {
			service.getVorgangAttachedItemServiceStub();

			verify(vorgangAttachedItemServiceBlockingStub).withInterceptors(any(NachrichtenCallContextAttachingInterceptor.class));
		}

		@Test
		void shouldReturnStub() {
			when(vorgangAttachedItemServiceBlockingStub.withInterceptors(any())).thenReturn(stubWithInterceptor);

			var result = service.getVorgangAttachedItemServiceStub();

			assertThat(result).isSameAs(stubWithInterceptor);
		}
	}

	@DisplayName("Find by vorgangId")
	@Nested
	class TestFindByVorgangId {

		private static final GrpcVorgangAttachedItem GRPC_VORGANG_ATTACHED_ITEM = GrpcVorgangAttachedItemTestFactory.create();

		@Mock
		private GrpcFindVorgangAttachedItemRequest grpcFindVorgangAttachedItemRequest;
		@Mock
		private GrpcFindVorgangAttachedItemResponse grpcFindVorgangAttachedItemResponse;

		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void init() {
			when(grpcFindVorgangAttachedItemResponse.getVorgangAttachedItemsList()).thenReturn(Collections.singletonList(GRPC_VORGANG_ATTACHED_ITEM));
			when(vorgangAttachedItemServiceBlockingStub.find(any())).thenReturn(grpcFindVorgangAttachedItemResponse);
			doReturn(vorgangAttachedItemServiceBlockingStub).when(service).getVorgangAttachedItemServiceStub();
			doReturn(grpcFindVorgangAttachedItemRequest).when(service).buildFindRequest(any());
		}

		@BeforeEach
		void mock() {
			when(postfachNachrichtMapper.fromAttachedItem(any())).thenReturn(postfachNachricht);
		}

		@Test
		void shouldGetServiceStub() {
			service.findByVorgangId(MessageTestFactory.VORGANG_ID).toList();

			verify(service).getVorgangAttachedItemServiceStub();
		}

		@Test
		void shouldCallServiceStub() {
			service.findByVorgangId(MessageTestFactory.VORGANG_ID).toList();

			verify(vorgangAttachedItemServiceBlockingStub).find(grpcFindVorgangAttachedItemRequest);
		}

		@Test
		void shouldCallMapper() {
			service.findByVorgangId(MessageTestFactory.VORGANG_ID).toList();

			verify(postfachNachrichtMapper).fromAttachedItem(GRPC_VORGANG_ATTACHED_ITEM);
		}

		@Test
		void shouldReturnMappedValue() {
			var postfachNachrichten = service.findByVorgangId(MessageTestFactory.VORGANG_ID);

			assertThat(postfachNachrichten).containsExactly(postfachNachricht);
		}
	}

	@Nested
	class TestGetPostfachNachrichtById {

		@Mock
		private VorgangAttachedItemServiceBlockingStub stubWithInterceptor;

		private final GrpcVorgangAttachedItemRequest request = GrpcVorgangAttachedItemRequestTestFactory.create();
		private final GrpcVorgangAttachedItemResponse response = GrpcVorgangAttachedItemResponseTestFactory.create();

		@BeforeEach
		void mock() {
			doReturn(stubWithInterceptor).when(service).getVorgangAttachedItemServiceStub();
			doReturn(request).when(service).buildGetByIdRequest(PostfachNachrichtTestFactory.ID);
			when(stubWithInterceptor.getById(request)).thenReturn(response);
		}

		@Test
		void shouldCallGetVorgangAttachedItemServiceStub() {
			callGetPostfachNachrichtById();

			verify(service).getVorgangAttachedItemServiceStub();
		}

		@Test
		void shouldCallBuildGetByIdRequest() {
			callGetPostfachNachrichtById();

			verify(service).buildGetByIdRequest(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldCallServiceStubWithInterceptor() {
			callGetPostfachNachrichtById();

			verify(stubWithInterceptor).getById(request);
		}

		@Test
		void shouldMapToPostfachNachricht() {
			callGetPostfachNachrichtById();

			verify(postfachNachrichtMapper).fromAttachedItem(GrpcVorgangAttachedItemResponseTestFactory.ATTACHED_ITEM);
		}

		@Test
		void shouldReturnMappedPostfachNachricht() {
			var mappedPostfachNachricht = PostfachNachrichtTestFactory.create();
			when(postfachNachrichtMapper.fromAttachedItem(GrpcVorgangAttachedItemResponseTestFactory.ATTACHED_ITEM))
					.thenReturn(mappedPostfachNachricht);

			var postfachNachricht = callGetPostfachNachrichtById();

			assertThat(postfachNachricht).isEqualTo(mappedPostfachNachricht);
		}

		private PostfachNachricht callGetPostfachNachrichtById() {
			return service.getPostfachNachrichtById(PostfachNachrichtTestFactory.ID);
		}
	}

	@Nested
	class TestBuildGetByIdRequest {

		@Test
		void shouldReturnGrpcRequest() {
			var request = service.buildGetByIdRequest(PostfachNachrichtTestFactory.ID);

			assertThat(request).isEqualTo(GrpcVorgangAttachedItemRequestTestFactory.create());
		}
	}
}