/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandId;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.nachrichten.attributes.ClientAttributeService;
import de.ozgcloud.nachrichten.common.vorgang.VorgangService;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.Direction;
import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;

class PersistPostfachNachrichtServiceImplTest {

	@Spy
	@InjectMocks
	private PersistPostfachNachrichtServiceImpl service;
	@Mock
	private VorgangService vorgangService;
	@Mock
	private OzgCloudCommandService commandService;
	@Mock
	private AttachedItemRemoteService attachedItemRemoteService;
	@Mock
	private PostfachNachrichtMapper postfachNachrichtMapper;
	@Mock
	private ClientAttributeService clientAttributeService;

	@DisplayName("Persist nachricht")
	@Nested
	class TestPersistNachricht {

		private final String userId = UUID.randomUUID().toString();
		private final Optional<String> userIdOpt = Optional.of(userId);

		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.createBuilder().id(null).build();

		private final OzgCloudCommand command = OzgCloudCommand.builder().build();

		@BeforeEach
		void mock() {
			doReturn(command).when(service).buildNachrichtCommand(any(), any());
		}

		@Test
		void shouldBuildNachrichtCommand() {
			persistNachricht();

			verify(service).buildNachrichtCommand(userIdOpt, nachricht);
		}

		@Test
		void shouldCallCommandService() {
			persistNachricht();

			verify(commandService).create(command);
		}

		@Test
		void shouldSetHasNewPostfachNachricht() {
			persistNachricht();

			verify(service).setHasNewPostfachNachricht(nachricht);
		}

		private void persistNachricht() {
			service.persistNachricht(userIdOpt, nachricht);
		}
	}

	@DisplayName("Create base command builder")
	@Nested
	class TestCreateBaseComandBuilder {

		private final String userId = UUID.randomUUID().toString();
		private final Optional<String> userIdOpt = Optional.of(userId);

		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.createBuilder().id(null).build();

		@Test
		void shouldContainVorgangId() {
			var command = service.createBaseCommandBuilder(userIdOpt, nachricht).build();

			assertThat(command.getVorgangId()).hasToString(MessageTestFactory.VORGANG_ID);
		}

		@DisplayName("created by")
		@Nested
		class TestCreatedBy {

			@Test
			void shouldBeAddIfIsPresent() {
				var command = service.createBaseCommandBuilder(userIdOpt, nachricht).build();

				assertThat(command.getCreatedBy()).hasToString(userId);
			}

			@Test
			void shouldNotBeAddIfMissing() {
				var command = service.createBaseCommandBuilder(Optional.empty(), nachricht).build();

				assertThat(command.getCreatedBy()).isNull();
			}
		}
	}

	@DisplayName("Build nachricht command")
	@Nested
	class TestBuildNachrichtCommand {

		private final String userId = UUID.randomUUID().toString();
		private final Optional<String> userIdOpt = Optional.of(userId);

		private final OzgCloudCommand command = OzgCloudCommand.builder().build();

		@DisplayName("on present id")
		@Nested
		class TestOnPresentId {

			private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.create();

			@BeforeEach
			void mock() {
				doReturn(command).when(service).buildUpdateCommand(any(), any());
			}

			@Test
			void shouldBuildUpdateCommand() {
				service.buildNachrichtCommand(userIdOpt, nachricht);

				verify(service).buildUpdateCommand(userIdOpt, nachricht);
			}

			@Test
			void shouldReturnValue() {
				var nachrichtCommand = service.buildNachrichtCommand(userIdOpt, nachricht);

				assertThat(nachrichtCommand).isEqualTo(command);
			}
		}

		@DisplayName("on missing id")
		@Nested
		class TestOnMissingId {

			private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.createBuilder().id(null).build();

			@BeforeEach
			void mock() {
				doReturn(command).when(service).buildCreateCommand(any(), any());
			}

			@Test
			void shouldBuildCreateCommand() {
				service.buildNachrichtCommand(userIdOpt, nachricht);

				verify(service).buildCreateCommand(userIdOpt, nachricht);
			}

			@Test
			void shouldReturnValue() {
				var nachrichtCommand = service.buildNachrichtCommand(userIdOpt, nachricht);

				assertThat(nachrichtCommand).isEqualTo(command);
			}
		}

		@DisplayName("for update")
		@Nested
		class TestBuildUpdateCommand {

			private final OzgCloudCommand.OzgCloudCommandBuilder commandBuilder = OzgCloudCommand.builder();

			private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.create();

			@BeforeEach
			void mock() {
				doReturn(commandBuilder).when(service).createBaseCommandBuilder(any(), any());
			}

			@Test
			void shouldCallCreateBaseCommandBuilder() {
				buildUpdateCommand();

				verify(service).createBaseCommandBuilder(userIdOpt, nachricht);
			}

			@Test
			void shouldContainOrder() {
				var command = buildUpdateCommand();

				assertThat(command.getOrder()).isEqualTo(PersistPostfachNachrichtServiceImpl.UPDATE_ATTACHED_ITEM_ORDER);
			}

			@Test
			void shouldContainRelationId() {
				var command = buildUpdateCommand();

				assertThat(command.getRelationId()).hasToString(PostfachNachrichtTestFactory.ID);
			}

			@DisplayName("bodyObject")
			@Nested
			class TestBodyObject {

				private final Map<String, Object> itemMap = Map.of("dummyKey", "dummyValue");

				@BeforeEach
				void mock() {
					when(postfachNachrichtMapper.mapToMap(any())).thenReturn(itemMap);
				}

				@Test
				void shouldContainClient() {
					var command = buildUpdateCommand();

					assertThat(command.getBodyObject()).containsEntry(PersistPostfachNachrichtServiceImpl.CLIENT_FIELD,
							PersistPostfachNachrichtServiceImpl.CLIENT);
				}

				@Test
				void shouldCallMapper() {
					buildUpdateCommand();

					verify(postfachNachrichtMapper).mapToMap(nachricht);
				}

				@Test
				void shouldContainItem() {
					var command = buildUpdateCommand();

					assertThat(command.getBodyObject()).containsEntry(PersistPostfachNachrichtServiceImpl.ITEM_FIELD, itemMap);
				}
			}

			private OzgCloudCommand buildUpdateCommand() {
				return service.buildUpdateCommand(userIdOpt, nachricht);
			}
		}
	}

	@DisplayName("Persist answer")
	@Nested
	class TestPersistAnswer {

		@Captor
		private ArgumentCaptor<PostfachNachricht> postfachNachrichtCaptor;

		private final OzgCloudCommand command = OzgCloudCommand.builder().id(OzgCloudCommandId.from(UUID.randomUUID().toString())).build();
		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.createBuilder().referencedNachricht(null).build();

		@BeforeEach
		void mock() {
			doReturn(command).when(service).buildCreateCommand(any(), any());
			when(commandService.create(any())).thenReturn(command);
		}

		@Test
		void shouldCallBuildCommand() {
			persistAnswer();

			verify(service).buildCreateCommand(eq(Optional.of(PostfachNachrichtTestFactory.CREATED_BY)), postfachNachrichtCaptor.capture());
			assertThat(postfachNachrichtCaptor.getValue().getReferencedNachricht()).isEqualTo(GrpcPostfachMailTestFactory.REFERENCED_NACHRICHT_ID);
		}

		@Test
		void shouldCallCommandService() {
			persistAnswer();

			verify(commandService).create(command);
		}

		@Test
		void shouldSetHasNewPostfachNachricht() {
			persistAnswer();

			verify(service).setHasNewPostfachNachricht(nachricht);
		}

		void persistAnswer() {
			service.persistAnswer(GrpcPostfachMailTestFactory.REFERENCED_NACHRICHT_ID, nachricht);
		}
	}

	@DisplayName("Build create command")
	@Nested
	class TestBuildCreateCommand {

		private final OzgCloudCommand.OzgCloudCommandBuilder commandBuilder = OzgCloudCommand.builder();

		private final String userId = UUID.randomUUID().toString();
		private final Optional<String> userIdOpt = Optional.of(userId);

		private final PostfachNachricht nachricht = PostfachNachrichtTestFactory.createBuilder().id(null).build();
		private final Map<String, Object> itemMap = Map.of("dummyKey", "dummyValue");

		@BeforeEach
		void mock() {
			doReturn(itemMap).when(service).buildCreateItem(any());
			doReturn(commandBuilder).when(service).createBaseCommandBuilder(any(), any());
		}

		@Test
		void shouldCallCreateBaseCommandBuilder() {
			buildCreateCommand();

			verify(service).createBaseCommandBuilder(userIdOpt, nachricht);
		}

		@Test
		void shouldContainOrder() {
			var command = buildCreateCommand();

			assertThat(command.getOrder()).isEqualTo(PersistPostfachNachrichtServiceImpl.CREATE_ATTACHED_ITEM_ORDER);
		}

		@Test
		void shouldContainRelationId() {
			var command = buildCreateCommand();

			assertThat(command.getRelationId()).hasToString(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldContainBodyObject() {
			var command = buildCreateCommand();

			assertThat(command.getBodyObject()).isEqualTo(itemMap);
		}

		private OzgCloudCommand buildCreateCommand() {
			return service.buildCreateCommand(userIdOpt, nachricht);
		}
	}

	@DisplayName("Build create item")
	@Nested
	class TestBuildCommandBody {

		private final PostfachNachricht postfachNachricht = PostfachNachrichtTestFactory.create();
		private final Map<String, Object> postfachNachrichtAsMap = Map.of("dummyKey", "dummyValue");

		@BeforeEach
		void mock() {
			when(postfachNachrichtMapper.mapToMap(any())).thenReturn(postfachNachrichtAsMap);
		}

		@Test
		void shouldHaveSetClient() {
			var body = service.buildCreateItem(postfachNachricht);

			assertThat(body).containsEntry(PersistPostfachNachrichtServiceImpl.CLIENT_FIELD, PersistPostfachNachrichtServiceImpl.CLIENT);
		}

		@Test
		void shouldHaveSetVorgangId() {
			var body = service.buildCreateItem(postfachNachricht);

			assertThat(body).containsEntry(PersistPostfachNachrichtServiceImpl.VORGANG_ID_FIELD, MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldHaveSetItemName() {
			var body = service.buildCreateItem(postfachNachricht);

			assertThat(body).containsEntry(PersistPostfachNachrichtServiceImpl.ITEM_NAME_FIELD, PersistPostfachNachrichtServiceImpl.ITEM_NAME);
		}

		@Test
		void shouldCallMapper() {
			service.buildCreateItem(postfachNachricht);

			verify(postfachNachrichtMapper).mapToMap(postfachNachricht);
		}

		@Test
		void shouldHaveSetItem() {
			var body = service.buildCreateItem(postfachNachricht);

			assertThat(body).containsEntry(PersistPostfachNachrichtServiceImpl.ITEM_FIELD, postfachNachrichtAsMap);
		}
	}

	@DisplayName("Set has new postfach nachricht")
	@Nested
	class TestSetHasNewPostfachNachricht {

		@Test
		void shouldCallClientAttributeServiceIfIsIncomingNachricht() {
			var incomingNachricht = PostfachNachrichtTestFactory.createBuilder().direction(Direction.IN).build();

			service.setHasNewPostfachNachricht(incomingNachricht);

			verify(clientAttributeService).setHasNewPostfachNachricht(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldNOTCallClientAttributeServiceIfIsOutgoindNachricht() {
			var outgoingNachricht = PostfachNachrichtTestFactory.createBuilder().direction(Direction.OUT).build();

			service.setHasNewPostfachNachricht(outgoingNachricht);

			verify(clientAttributeService, never()).setHasNewPostfachNachricht(any());
		}
	}

	@DisplayName("Find answers")
	@Nested
	class TestFindAnswers {

		private final String serviceKontoType = "serviceKontoTypeDummy";

		private final PostfachNachricht postfachNachrichtWithNoReferencedNachricht = PostfachNachrichtTestFactory.createBuilder()
				.referencedNachricht(null).build();
		private final PostfachNachricht postfachNachrichtWithNonMatchingReferencedNachricht = PostfachNachrichtTestFactory.createBuilder()
				.referencedNachricht("quatsch").build();
		private final PostfachNachricht matchingPostfachNachricht = PostfachNachrichtTestFactory.create();

		@BeforeEach
		void mock() {
			when(vorgangService.findVorgangIds(any(), any())).thenReturn(Stream.of(MessageTestFactory.VORGANG_ID));
			when(attachedItemRemoteService.findByVorgangId(any())).thenReturn(Stream.of(postfachNachrichtWithNoReferencedNachricht,
					matchingPostfachNachricht, postfachNachrichtWithNonMatchingReferencedNachricht));
		}

		@Test
		void shouldCallVorgangRemoteService() {
			doFindAnswers().toList();

			verify(vorgangService).findVorgangIds(serviceKontoType, StringBasedIdentifierTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE);
		}

		@Test
		void shouldCallAttachedItemRemoteService() {
			doFindAnswers().toList();

			verify(attachedItemRemoteService).findByVorgangId(MessageTestFactory.VORGANG_ID);
		}

		@Test
		void shouldReturnMatchingPostfachNachricht() {
			var answers = doFindAnswers();

			assertThat(answers).containsExactly(matchingPostfachNachricht);
		}

		private Stream<PostfachNachricht> doFindAnswers() {
			return service.findAnswers(serviceKontoType, StringBasedIdentifierTestFactory.STRING_BASED_IDENTIFIER_POSTFACH_ID_VALUE,
					GrpcPostfachMailTestFactory.REFERENCED_NACHRICHT_ID);
		}
	}
}
