/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.stream.Stream;

import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import io.grpc.stub.StreamObserver;

class PostfachGrpcServiceTest {

	@InjectMocks
	private PostfachGrpcService service;

	@Mock
	private PostfachService postfachService;
	@Mock
	private GrpcPostfachNachrichtMapper nachrichtMapper;
	@Mock
	private PostfachMapper postfachMapper;

	@Spy
	private PostfachNachrichtMapper mapper = Mappers.getMapper(PostfachNachrichtMapper.class);

	@Nested
	class TestSaveNachrichtDraft {

		@Mock
		private StreamObserver<GrpcSaveNachrichtDraftResponse> responseObserver;

		@BeforeEach
		void init() {
			when(nachrichtMapper.mapFromGrpc(any())).thenReturn(PostfachNachrichtTestFactory.create());
		}

		@Test
		void shouldCallService() {
			var nachricht = PostfachNachrichtTestFactory.create();
			when(nachrichtMapper.mapFromGrpc(any())).thenReturn(nachricht);

			service.saveNachrichtDraft(GrpcSaveNachrichtDraftRequestTestFactory.create(), responseObserver);

			verify(postfachService).saveDraft(GrpcPostfachMailTestFactory.VORGANG_ID, nachricht);
		}

		@Test
		void shouldSendResponse() {
			service.saveNachrichtDraft(GrpcSaveNachrichtDraftRequestTestFactory.create(), responseObserver);

			verify(responseObserver).onNext(any());
		}

		@Test
		void shouldCompleteCall() {
			service.saveNachrichtDraft(GrpcSaveNachrichtDraftRequestTestFactory.create(), responseObserver);

			verify(responseObserver).onCompleted();
		}

	}

	@Nested
	class TestSendMail {

		@Mock
		private StreamObserver<GrpcSendPostfachMailResponse> responseObserver;

		@Captor
		private ArgumentCaptor<PostfachNachricht> mailCaptor;

		@Test
		void shouldCallPostfachServiceWithCommandId() {
			service.sendPostfachMail(GrpcSendPostfachMailRequestTestFactory.create(), responseObserver);

			verify(postfachService).sendMail(eq(GrpcSendPostfachMailRequestTestFactory.COMMAND_ID), anyString(), any());
		}

		@Test
		void shouldCallServiceWithMail() {
			var grpcSendPostfachMailRequest = GrpcSendPostfachMailRequestTestFactory.create();

			service.sendPostfachMail(grpcSendPostfachMailRequest, responseObserver);

			verify(postfachService).sendMail(anyString(), anyString(), mailCaptor.capture());
			Assertions.assertThat(mailCaptor.getValue()).usingRecursiveComparison()
					.ignoringFields(PostfachNachricht.FIELD_MESSAGE_ID, PostfachNachricht.FIELD_CREATED_AT, PostfachNachricht.FIELD_CREATED_BY,
							PostfachNachricht.FIELD_DIRECTION, PostfachNachricht.FIELD_SENT_SUCCESSFUL, PostfachNachricht.FIELD_SENT_AT,
							PostfachNachricht.FIELD_MESSAGE_CODE,
							PostfachNachricht.FIELD_ATTACHMENTS) // TODO remove this for sending attachments
					.isEqualTo(createPostfachNachricht());
		}

		private PostfachNachricht createPostfachNachricht() {
			return PostfachNachrichtTestFactory.createBuilder()
					.id(GrpcPostfachMailTestFactory.ID)
					.vorgangId(GrpcPostfachMailTestFactory.VORGANG_ID)
					.postfachAddress(PostfachAddressTestFactory.createBuilder()
							.identifier(StringBasedIdentifier.builder().postfachId(GrpcPostfachMailTestFactory.POSTFACH_ID).build()).build())
					.build();
		}

		@Test
		void shouldCallServiceWithUserId() {
			service.sendPostfachMail(GrpcSendPostfachMailRequestTestFactory.create(), responseObserver);

			verify(postfachService).sendMail(anyString(), eq(GrpcSendPostfachMailRequestTestFactory.USER_ID), any());
		}

		@Test
		void shouldSendResponse() {
			service.sendPostfachMail(GrpcSendPostfachMailRequestTestFactory.create(), responseObserver);

			verify(responseObserver).onNext(notNull());
		}

		@Test
		void shouldCompleteCall() {
			service.sendPostfachMail(GrpcSendPostfachMailRequestTestFactory.create(), responseObserver);

			verify(responseObserver).onCompleted();
		}
	}

	@Nested
	class TestFindPostfachMail {

		private GrpcFindPostfachMailRequest request;
		@Mock
		private StreamObserver<GrpcFindPostfachMailResponse> responseObserver;

		@BeforeEach
		void init() {
			request = GrpcFindPostfachMailRequest.newBuilder().setNachrichtId(PostfachNachrichtTestFactory.ID).build();
		}

		@Test
		void shouldCallService() {
			service.findPostfachMail(request, responseObserver);

			verify(postfachService).findById(PostfachNachrichtTestFactory.ID);
		}

		@Test
		void shouldCallOnNext() {
			service.findPostfachMail(request, responseObserver);

			verify(responseObserver).onNext(notNull());
		}

		@Test
		void shouldCompleteCall() {
			service.findPostfachMail(request, responseObserver);

			verify(responseObserver).onCompleted();
		}
	}

	@Nested
	class TestFindPostfachMails {

		private static final String VORGANG_ID = "42";
		private GrpcFindPostfachMailsRequest request;

		@Mock
		private StreamObserver<GrpcFindPostfachMailsResponse> responseObserver;

		@BeforeEach
		void init() {
			request = GrpcFindPostfachMailsRequest.newBuilder().setVorgangId(VORGANG_ID).build();
		}

		@Test
		void shouldCallService() {
			service.findPostfachMails(request, responseObserver);

			verify(postfachService).findByVorgang(VORGANG_ID);
		}

		@Test
		void shouldCompleteCall() {
			service.findPostfachMails(request, responseObserver);

			verify(responseObserver).onCompleted();
		}

	}

	@Nested
	class TestResendMail {

		@Mock
		private StreamObserver<GrpcResendPostfachMailResponse> responseObserver;

		@Test
		void shouldCallPostfachService() {
			service.resendPostfachMail(GrpcResendPostfachMailRequestTestFactory.create(), responseObserver);

			verify(postfachService).resendMail(GrpcResendPostfachMailRequestTestFactory.COMMAND_ID, GrpcPostfachMailTestFactory.ID);
		}

		@Test
		void shouldSendResponse() {
			service.resendPostfachMail(GrpcResendPostfachMailRequestTestFactory.create(), responseObserver);

			verify(responseObserver).onNext(notNull());
		}

		@Test
		void shouldCompleteCall() {
			service.resendPostfachMail(GrpcResendPostfachMailRequestTestFactory.create(), responseObserver);

			verify(responseObserver).onCompleted();
		}
	}

	@Nested
	class TestIsPostfachConfigured {

		private final GrpcIsPostfachConfiguredRequest request = GrpcIsPostfachConfiguredRequest.newBuilder().build();

		@Mock
		private StreamObserver<GrpcIsPostfachConfiguredResponse> responseObserver;

		@Test
		void shouldCallPostfachService() {
			service.isPostfachConfigured(request, responseObserver);

			verify(postfachService).isPostfachConfigured();
		}

		@Test
		void shouldSendResponse() {
			service.isPostfachConfigured(request, responseObserver);

			verify(responseObserver).onNext(notNull());
		}

		@Test
		void shouldCompleteCall() {
			service.isPostfachConfigured(request, responseObserver);

			verify(responseObserver).onCompleted();
		}
	}

	@Nested
	class TestGetPostfachConfig {
		private final GrpcGetPostfachConfigRequest request = GrpcGetPostfachConfigRequest.newBuilder().build();

		@Mock
		private StreamObserver<GrpcGetPostfachConfigResponse> responseObserver;

		@Test
		void shouldSendResponse() {
			service.getPostfachConfig(request, responseObserver);

			verify(responseObserver).onNext(notNull());
		}

		@Test
		void shouldCompleteCall() {
			service.getPostfachConfig(request, responseObserver);

			verify(responseObserver).onCompleted();
		}

		@Nested
		class TestBuildGetPostfachConfigResponse {

			@Test
			void shouldReturnNotConfigured() {
				var response = service.buildGetPostfachConfigResponse();

				assertThat(response).isEqualTo(GrpcGetPostfachConfigResponse.newBuilder().setConfigured(false).build());
			}

			@Nested
			class TestConfigured {

				@Mock
				private Postfach postfach;
				@Mock
				private GrpcPostfach grpcPostfach;

				@BeforeEach
				void setup() {
					when(postfachService.getPostfachs()).thenReturn(Stream.of(postfach));
					when(postfachMapper.toGrpc(postfach)).thenReturn(grpcPostfach);
				}

				@Test
				void shouldReturnPostfachs() {
					var response = service.buildGetPostfachConfigResponse();

					assertThat(response.getConfigured()).isTrue();
				}

				@Test
				void shouldSetPostfach() {
					var response = service.buildGetPostfachConfigResponse();

					assertThat(response.getPostfachList()).containsExactly(grpcPostfach);
				}
			}
		}
	}
}