/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.UUID;

import org.apache.commons.lang3.StringUtils;
import org.assertj.core.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory;
import de.ozgcloud.vorgang.common.GrpcObject;
import de.ozgcloud.vorgang.common.grpc.GrpcObjectMapper;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItem;
import lombok.SneakyThrows;

class PostfachNachrichtMapperTest {

	@Spy
	@InjectMocks
	private PostfachNachrichtMapper mapper = Mappers.getMapper(PostfachNachrichtMapper.class);
	@Mock
	private GrpcObjectMapper grpcObjectMapper;

	@DisplayName("From map")
	@Nested
	class TestFromMap {

		@DisplayName("with existing postfachAddress")
		@Nested
		class TestWithExistingPostfachAddress {

			@BeforeEach
			void mockGrpcObjectMapper() {
				when(grpcObjectMapper.fromMap(any())).thenReturn(GrpcPostfachAddressTestFactory.createIdentifier());
			}

			@Test
			void shouldFillGrpcMail() {
				var grpcMail = fromMap(createPostfachNachrichtAsMap());

				assertThat(grpcMail).usingRecursiveComparison().isEqualTo(GrpcPostfachMailTestFactory.create());
			}

			private Map<String, Object> createPostfachNachrichtAsMap() {
				var map = PostfachNachrichtTestFactory.asMap();
				map.put(PostfachNachricht.FIELD_CREATED_BY, GrpcPostfachMailTestFactory.CREATED_BY);
				map.put(PostfachNachricht.FIELD_ID, GrpcPostfachMailTestFactory.ID);
				map.put(PostfachNachricht.FIELD_MESSAGE_CODE, GrpcPostfachMailTestFactory.MESSAGE_CODE);
				map.put(PostfachAddress.FIELD_POSTFACH_ID, GrpcPostfachMailTestFactory.POSTFACH_ID);
				map.put(PostfachNachricht.FIELD_VORGANG_ID, GrpcPostfachMailTestFactory.VORGANG_ID);
				return map;
			}

			@Test
			void shouldCallGrpcObjectMapper() {
				fromMap(getMapWithoutProperty(PostfachNachricht.FIELD_SENT_SUCCESSFUL));

				verify(grpcObjectMapper).fromMap(any());
			}

			@Test
			void shouldIgnoreSentSuccessful() {
				var grpcMail = fromMap(getMapWithoutProperty(PostfachNachricht.FIELD_SENT_SUCCESSFUL));

				assertThat(grpcMail.getSentSuccessful()).isFalse();
			}

			@Test
			void shouldIgnoreSentAt() {
				var grpcMail = fromMap(getMapWithoutProperty(PostfachNachricht.FIELD_SENT_AT));

				assertThat(grpcMail.getSentAt()).isEmpty();
			}

			@Test
			void shouldProceedWithSingleAttachment() {
				var postfachNachrichtMap = PostfachNachrichtTestFactory.asMap();

				postfachNachrichtMap.put(PostfachNachricht.FIELD_ATTACHMENTS, UUID.randomUUID().toString());

				assertDoesNotThrow(() -> fromMap(postfachNachrichtMap));
			}

			@Test
			void shouldProceedWithNullAsAttachments() {
				var postfachNachrichtMap = PostfachNachrichtTestFactory.asMap();

				postfachNachrichtMap.put(PostfachNachricht.FIELD_ATTACHMENTS, null);

				assertDoesNotThrow(() -> fromMap(postfachNachrichtMap));
			}

			@DisplayName("on empty attachments")
			@Nested
			class TestOnEmptyAttachmets {

				@Test
				void shouldProceedWithEmptyAttachments() {
					var postfachNachrichtMap = PostfachNachrichtTestFactory.asMap();

					postfachNachrichtMap.put(PostfachNachricht.FIELD_ATTACHMENTS, StringUtils.EMPTY);

					assertDoesNotThrow(() -> fromMap(postfachNachrichtMap));
				}

				@Test
				void shouldKeepEmptyAttachments() {
					var postfachNachrichtMap = PostfachNachrichtTestFactory.asMap();
					postfachNachrichtMap.put(PostfachNachricht.FIELD_ATTACHMENTS, StringUtils.EMPTY);

					var postfachNachricht = fromMap(postfachNachrichtMap);

					assertThat(postfachNachricht.getAttachmentList()).isEmpty();
				}
			}
		}

		@DisplayName("with missing postfachAddress")
		@Nested
		class TestWithMissinPostfachAddress {

			@Test
			void shouldNotThrowException() {
				assertDoesNotThrow(() -> fromMap(getMapWithoutProperty(PostfachNachricht.POSTFACH_ADDRESS_FIELD)));
			}
		}

		private GrpcPostfachMail fromMap(Map<String, Object> map) {
			return mapper.fromMap(map);
		}
	}

	@DisplayName("From map to postfachMail")
	@Nested
	class TestFromMapToPostfachMail {

		@Test
		void shouldMap() {
			var mail = fromMapToPostfachMail(PostfachNachrichtTestFactory.asMap());

			Assertions.assertThat(mail).usingRecursiveComparison().isEqualTo(PostfachNachrichtTestFactory.create());
		}

		@Test
		void shouldMapSingleAttachment() {
			var mailMap = PostfachNachrichtTestFactory.asMap();
			mailMap.put(PostfachNachricht.FIELD_ATTACHMENTS, PostfachNachrichtTestFactory.ATTACHMENT_FILE_ID);

			var mail = fromMapToPostfachMail(mailMap);

			assertThat(mail.getAttachments()).containsExactly(PostfachNachrichtTestFactory.ATTACHMENT_FILE_ID);
		}

		@DisplayName("with missing postfachAddress")
		@Nested
		class TestWithMissingPostfachAddress {

			@Test
			void shouldMapAsNull() {
				var mail = fromMapToPostfachMail(getMapWithoutProperty(PostfachNachricht.POSTFACH_ADDRESS_FIELD));

				Assertions.assertThat(mail.getPostfachAddress()).isNull();
			}

			@Test
			void shouldNotThrowExceptionOnMissingPostfachAddress() {
				assertDoesNotThrow(() -> fromMapToPostfachMail(getMapWithoutProperty(PostfachNachricht.POSTFACH_ADDRESS_FIELD)));
			}
		}

		private PostfachNachricht fromMapToPostfachMail(Map<String, Object> map) {
			return mapper.fromMapToPostfachMail(map);
		}
	}

	@Nested
	class TestFromAttachedItem {

		@Mock
		private GrpcVorgangAttachedItem grpcVorgangAttachedItem;
		@Mock
		private GrpcObject grpcObject;

		private final Map<String, Object> itemMap = PostfachNachrichtTestFactory.asMap();

		@BeforeEach
		void init() {
			when(grpcVorgangAttachedItem.getItem()).thenReturn(grpcObject);
			when(grpcObjectMapper.mapFromGrpc(any())).thenReturn(itemMap);
		}

		@Test
		void shouldCallGrpcObjectMapper() {
			mapper.fromAttachedItem(grpcVorgangAttachedItem);

			verify(grpcObjectMapper).mapFromGrpc(grpcObject);
		}

		@Test
		void shouldCallFromMapToPostfachMail() {
			mapper.fromAttachedItem(grpcVorgangAttachedItem);

			verify(mapper).fromMapToPostfachMail(itemMap);
		}

		@Test
		void shouldReturnMappedNachricht() {
			var postfachNachricht = PostfachNachrichtTestFactory.create();
			doReturn(postfachNachricht).when(mapper).fromMapToPostfachMail(itemMap);

			var mapped = mapper.fromAttachedItem(grpcVorgangAttachedItem);

			assertThat(mapped).isEqualTo(postfachNachricht);
		}
	}

	@DisplayName("From grpc mail")
	@Nested
	class TestFromGrpcMail {

		@Test
		void shouldMapFields() {
			var mapped = fromGrpcMail(createPostfachMail());

			Assertions.assertThat(mapped).usingRecursiveComparison()
					.ignoringFields("createdAt", "createdBy", "direction", "messageId", "sentAt", "sentSuccessful", "memoizedHashCode")
					.isEqualTo(PostfachNachrichtTestFactory.create());
		}

		private GrpcPostfachMail createPostfachMail() {
			return GrpcPostfachMailTestFactory.createBuilder()
					.setId(PostfachNachrichtTestFactory.ID)
					.setMessageCode(PostfachNachrichtTestFactory.MESSAGE_CODE)
					.setPostfachAddress(GrpcPostfachAddressTestFactory.createBuilder()
							.setIdentifier(GrpcPostfachAddressTestFactory.createIdentifier(MessageTestFactory.POSTFACH_ID)))
					.setVorgangId(MessageTestFactory.VORGANG_ID).build();
		}

		@Test
		void shouldHandleEmptyPostfachId() {
			var postfachMailWithoutPostfachId = GrpcPostfachMailTestFactory.createBuilder().build();

			assertDoesNotThrow(() -> fromGrpcMail(postfachMailWithoutPostfachId));
		}

		private PostfachNachricht fromGrpcMail(GrpcPostfachMail postfachMail) {
			return mapper.fromGrpcMail(postfachMail);
		}
	}

	private Map<String, Object> getMapWithoutProperty(String entryKey) {
		var map = PostfachNachrichtTestFactory.asMap();
		map.remove(entryKey);
		return map;
	}

	private Object getAttributeFromGrpcObject(Object obj, String name) {
		return getAttributeFrom(obj, name + "_");
	}

	@SneakyThrows
	private Object getAttributeFromObject(Object obj, String name) {
		return getAttributeFrom(obj, name);
	}

	@SneakyThrows
	private Object getAttributeFrom(Object obj, String name) {
		var clazz = obj.getClass();
		var field = clazz.getDeclaredField(name);
		field.setAccessible(true);

		return field.get(obj);
	}

	@DisplayName("Map to map")
	@Nested
	class TestMapToMap {

		@Test
		void shouldContainAllFieldsIgnoringId() {
			var map = mapper.mapToMap(PostfachNachrichtTestFactory.createBuilder().id(null).build());

			var mapWithoutId = PostfachNachrichtTestFactory.asMap();
			mapWithoutId.remove("id");

			assertThat(map).containsAllEntriesOf(mapWithoutId);
		}

		@Test
		void shouldIgnoreEmptyMessageId() {
			assertDoesNotThrow(() -> mapper.mapToMap(PostfachNachrichtTestFactory.createBuilder().messageId(null).build()));
		}

		@Test
		void shouldIgnoreEmptyCreatedBy() {
			assertDoesNotThrow(() -> mapper.mapToMap(PostfachNachrichtTestFactory.createBuilder().createdBy(null).build()));
		}

		@Test
		void shouldHandleNullAsPostfachId() {
			var postfachMail = PostfachNachrichtTestFactory.create();

			assertDoesNotThrow(() -> mapper.mapToMap(postfachMail));
		}

		@Test
		void shouldProceedWithEmptyPostfachAddress() {
			var postfachMail = PostfachNachrichtTestFactory.createBuilder().postfachAddress(null).build();

			assertDoesNotThrow(() -> mapper.mapToMap(postfachMail));
		}

		@Test
		void shouldHandleEmptyReferencedNachrichtId() {
			var postfachNachricht = PostfachNachrichtTestFactory.createBuilder().id(null)
					.referencedNachricht(null)
					.build();

			assertDoesNotThrow(() -> mapper.mapToMap(postfachNachricht));
		}
	}
}
