/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.BiFunction;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.nachrichten.postfach.bayernid.FileSender.StreamReader;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageRequest;
import io.grpc.stub.CallStreamObserver;
import lombok.SneakyThrows;

class FileSenderTest {

	private FileSender fileSender;

	@Mock
	private BiFunction<byte[], Integer, GrpcSendBayernIdMessageRequest> chunkBuilder;
	@Mock
	private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;
	@Mock
	private GrpcSendBayernIdMessageRequest metadata;
	@Mock
	private InputStream inputStream;

	@BeforeEach
	void setup() {
		fileSender = spy(new FileSender(chunkBuilder, requestObserver, metadata, inputStream));
	}

	@Nested
	class TestSend {

		@Test
		void shouldCallSendMetadata() {
			fileSender.send();

			verify(fileSender).sendMetadata();
		}

		@Test
		void shouldCallSendNextChunk() {
			when(requestObserver.isReady()).thenReturn(true);
			doAnswer(invocation -> {
				((AtomicBoolean) ReflectionTestUtils.getField(fileSender, "done")).set(true);
				return null;
			}).when(fileSender).sendNextChunk();

			fileSender.send();

			verify(fileSender).sendNextChunk();
		}

		@Test
		void shouldNotCallSendNextChunkWhenDone() {
			((AtomicBoolean) ReflectionTestUtils.getField(fileSender, "done")).set(true);

			fileSender.send();

			verify(fileSender, never()).sendNextChunk();
		}

		@Test
		void shouldNotCallSendNextChunkWhenNotReady() {
			fileSender.send();

			verify(fileSender, never()).sendNextChunk();
		}
	}

	@Nested
	class TestSendMetadata {

		@Test
		void shouldCallOnNextOnce() {
			fileSender.sendMetadata();
			fileSender.sendMetadata();

			verify(requestObserver).onNext(metadata);
		}

	}

	@Nested
	class TestSendNextChunk {

		@Test
		void shouldCallSendChunk() {
			var contentToSend = new byte[] { 1, 2, 3, 4, 5 };
			setContent(contentToSend);

			fileSender.sendNextChunk();

			verify(fileSender).sendChunk(contentToSend, 5);
		}

		@Test
		void shouldReturnContentLength() {
			var contentToSend = new byte[] { 1, 2, 3, 4, 5 };
			setContent(contentToSend);

			var length = fileSender.sendNextChunk();

			assertThat(length).isEqualTo(contentToSend.length);
		}

		@Test
		void shouldCallEndTransfer() {
			setContent(new byte[] {});

			fileSender.sendNextChunk();

			verify(fileSender).endTransfer();
		}

		@SneakyThrows
		void setContent(byte[] contentToSend) {
			var streamReader = getStreamReader();
			ReflectionTestUtils.setField(streamReader, "buffer", contentToSend);
			when(inputStream.read(any(), anyInt(), anyInt())).thenReturn(contentToSend.length);
		}

	}

	@Nested
	class TestEndTransfer {

		@Mock
		private StreamReader streamReader;

		@BeforeEach
		void setup() {
			ReflectionTestUtils.setField(fileSender, "streamReader", streamReader);
		}

		@Test
		void shouldSetDone() {
			fileSender.endTransfer();

			var done = (AtomicBoolean) ReflectionTestUtils.getField(fileSender, "done");
			assertThat(done.get()).isTrue();
		}

		@Test
		void shouldCloseStream() {
			fileSender.endTransfer();

			verify(streamReader).close();
		}
	}

	@Nested
	class TestSendChunk {

		@Mock
		private GrpcSendBayernIdMessageRequest chunkRequest;

		private byte[] contentToSend = new byte[] { 1, 2, 3, 4, 5 };
		private int length = 5;


		@Test
		void shouldCallChunkBuilder() {
			fileSender.sendChunk(contentToSend, length);

			verify(chunkBuilder).apply(contentToSend, length);
		}

		@Test
		void shouldCallOnNext() {
			when(chunkBuilder.apply(any(), anyInt())).thenReturn(chunkRequest);

			fileSender.sendChunk(contentToSend, length);

			verify(requestObserver).onNext(chunkRequest);
		}

	}

	StreamReader getStreamReader() {
		return (StreamReader) ReflectionTestUtils.getField(fileSender, "streamReader");
	}
}