/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.InputStream;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.BiFunction;
import java.util.function.Function;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.common.binaryfile.BinaryFileUploadStreamObserver;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageRequest;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.GrpcSendBayernIdMessageResponse;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;

class MessageWithFilesSenderTest {

	private MessageWithFilesSender messageWithFilesSender;

	@Mock
	private Function<StreamObserver<GrpcSendBayernIdMessageResponse>, CallStreamObserver<GrpcSendBayernIdMessageRequest>> reqObserverBuilder;
	@Mock
	private GrpcSendBayernIdMessageRequest messageMetadata;
	@Mock
	private List<String> attachmentIds;
	@Mock
	private Function<String, BayernIdAttachment> toAttachment;
	@Mock
	private Function<BayernIdAttachment, GrpcSendBayernIdMessageRequest> attachmentMetadataMapper;
	@Mock
	private BiFunction<byte[], Integer, GrpcSendBayernIdMessageRequest> chunkBuilder;

	@BeforeEach
	void setup() {
		messageWithFilesSender = spy(
				MessageWithFilesSender.builder().reqObserverBuilder(reqObserverBuilder).messageMetadata(messageMetadata).toAttachment(toAttachment)
						.attachmentMetadataMapper(attachmentMetadataMapper).chunkBuilder(chunkBuilder).build());
	}

	@Nested
	class TestSend {

		@Mock
		private BinaryFileUploadStreamObserver<GrpcSendBayernIdMessageRequest, GrpcSendBayernIdMessageResponse> responseStreamObserver;
		@Mock
		private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;

		@Test
		void shouldReturnSameObject() {
			var result = messageWithFilesSender.send();

			assertThat(result).isSameAs(messageWithFilesSender);
		}

		@Test
		void shouldCallRequestObserverBuilder() {
			try (var createMock = Mockito.mockStatic(BinaryFileUploadStreamObserver.class)) {
				createMock.when(() -> BinaryFileUploadStreamObserver.create(any(), any())).thenReturn(responseStreamObserver);

				messageWithFilesSender.send();

				verify(reqObserverBuilder).apply(responseStreamObserver);
			}
		}

		@Test
		void shouldSetRequestStreamObserver() {
			when(reqObserverBuilder.apply(any())).thenReturn(requestObserver);

			messageWithFilesSender.send();

			assertThat(messageWithFilesSender).extracting("requestObserver").isSameAs(requestObserver);
		}
	}

	@Nested
	class TestSendNext {

		@Mock
		private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;

		@BeforeEach
		void setup() {
			ReflectionTestUtils.setField(messageWithFilesSender, "requestObserver", requestObserver);
		}

		@Test
		void shouldCheckIfDone() {
			doNothing().when(messageWithFilesSender).waitForObserver();

			messageWithFilesSender.sendNext();

			verify(messageWithFilesSender).waitForObserver();
		}

		@Test
		void shouldRetrunIfDone() {
			((AtomicBoolean) ReflectionTestUtils.getField(messageWithFilesSender, "done")).set(true);

			messageWithFilesSender.sendNext();

			verify(messageWithFilesSender, never()).waitForObserver();
		}

		@Test
		void shouldCallSendMetadata() {
			doNothing().when(messageWithFilesSender).waitForObserver();

			messageWithFilesSender.sendNext();

			verify(messageWithFilesSender).sendMetadata();
		}

		@Test
		void shouldCallSendAttachments() {
			doNothing().when(messageWithFilesSender).waitForObserver();

			messageWithFilesSender.sendNext();

			verify(messageWithFilesSender).sendAttachments();
		}
	}

	@Nested
	class TestSendMetadata {

		@Mock
		private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;

		@BeforeEach
		void setup() {
			ReflectionTestUtils.setField(messageWithFilesSender, "requestObserver", requestObserver);
		}

		@Test
		void shouldCallOnNextOnce() {
			messageWithFilesSender.sendMetadata();
			messageWithFilesSender.sendMetadata();

			verify(requestObserver).onNext(messageMetadata);
		}

	}

	@Nested
	class TestSendAttachments {

		@Test
		void shouldCallCompleteRequest() {
			doNothing().when(messageWithFilesSender).completeRequest();

			messageWithFilesSender.sendAttachments();

			verify(messageWithFilesSender).completeRequest();
		}

		@Test
		void shouldCallSendData() {
			ReflectionTestUtils.setField(messageWithFilesSender, "attachmentIds", List.of("attachmentId"));
			doNothing().when(messageWithFilesSender).sendData();

			messageWithFilesSender.sendAttachments();

			verify(messageWithFilesSender).sendData();
		}

	}

	@Nested
	class TestSendData {

		@Mock
		private FileSender fileSender;
		@Mock
		private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;

		@BeforeEach
		void setup() {
			ReflectionTestUtils.setField(messageWithFilesSender, "requestObserver", requestObserver);
		}

		@Test
		void shouldCallGetFileSender() {
			doReturn(Optional.of(fileSender)).when(messageWithFilesSender).getFileSender();

			messageWithFilesSender.sendData();

			verify(messageWithFilesSender).getFileSender();
		}

		@Test
		void shouldCallSendOnFileSender() {
			doReturn(Optional.of(fileSender)).when(messageWithFilesSender).getFileSender();
			when(requestObserver.isReady()).thenReturn(true, false);

			messageWithFilesSender.sendData();

			verify(fileSender).send();
		}

		@Test
		@DisplayName("should not complete request if attachments not sent")
		void shouldNotCompleteIfNotDone() {
			doReturn(Optional.of(fileSender)).when(messageWithFilesSender).getFileSender();
			when(requestObserver.isReady()).thenReturn(true, false);

			messageWithFilesSender.sendData();

			verify(messageWithFilesSender, never()).completeRequest();
		}

		@Test
		void shouldCompleteRequest() {
			doReturn(Optional.empty()).when(messageWithFilesSender).getFileSender();

			messageWithFilesSender.sendData();

			verify(messageWithFilesSender).completeRequest();
		}
	}

	@Nested
	class TestGetFileSender {

		@Mock
		private FileSender fileSender;

		@Nested
		class TestCreateFirstFileSender {

			@BeforeEach
			void setup() {
				doReturn(fileSender).when(messageWithFilesSender).buildNextFileSender();
			}

			@Test
			void shouldCallBuildNextSender() {
				messageWithFilesSender.getFileSender();

				verify(messageWithFilesSender).buildNextFileSender();
			}

			@Test
			void shouldSetCurrentFileSender() {
				messageWithFilesSender.getFileSender();

				assertThat(messageWithFilesSender).extracting("currentFileSender").isSameAs(fileSender);
			}

			@Test
			void shouldReturnCurrentFileSender() {
				var result = messageWithFilesSender.getFileSender();

				assertThat(result).containsSame(fileSender);
			}

		}

		@Nested
		class TestReturnCurrentFileSender {

			@BeforeEach
			void setup() {
				ReflectionTestUtils.setField(messageWithFilesSender, "currentFileSender", fileSender);
			}

			@Test
			void shouldReturnCurrentFileSender() {
				var result = messageWithFilesSender.getFileSender();

				assertThat(result).containsSame(fileSender);
			}

			@Test
			void shouldNotCallBuildNextSender() {
				messageWithFilesSender.getFileSender();

				verify(messageWithFilesSender, never()).buildNextFileSender();
			}
		}

		@Nested
		class TestCreateNextFileSender {

			@Mock
			private FileSender nextFileSender;

			@BeforeEach
			void setup() {
				when(fileSender.isDone()).thenReturn(true);
				ReflectionTestUtils.setField(messageWithFilesSender, "currentFileSender", fileSender);
				ReflectionTestUtils.setField(messageWithFilesSender, "attachmentIds", List.of("attachmentId"));
				doReturn(nextFileSender).when(messageWithFilesSender).buildNextFileSender();
			}

			@Test
			void shouldCallBuildNextSender() {
				messageWithFilesSender.getFileSender();

				verify(messageWithFilesSender).buildNextFileSender();
			}

			@Test
			void shouldSetCurrentFileSender() {
				messageWithFilesSender.getFileSender();

				assertThat(messageWithFilesSender).extracting("currentFileSender").isSameAs(nextFileSender);
			}

			@Test
			void shouldReturnCurrentFileSender() {
				var result = messageWithFilesSender.getFileSender();

				assertThat(result).containsSame(nextFileSender);
			}
		}

	}

	@Nested
	class TestBuildNextFileSender {

		private final String attachmentId = "id1";

		@Mock
		private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;
		@Mock
		private GrpcSendBayernIdMessageRequest attachmentMetadata;

		@Mock
		private InputStream inputStream;

		private final BayernIdAttachment attachment = BayernIdAttachmentTestFactory.create();

		@BeforeEach
		void setup() {
			ReflectionTestUtils.setField(messageWithFilesSender, "attachmentIds", List.of(attachmentId));
			when(toAttachment.apply(any())).thenReturn(attachment);
		}

		@Test
		void shouldCallToAttachment() {
			messageWithFilesSender.buildNextFileSender();

			verify(toAttachment).apply(attachmentId);
		}

		@Test
		void shouldSetChunkBuilder() {
			var result = messageWithFilesSender.buildNextFileSender();

			assertThat(result).extracting("chunkBuilder").isSameAs(chunkBuilder);
		}

		@Test
		void shouldSetRequestObserver() {
			ReflectionTestUtils.setField(messageWithFilesSender, "requestObserver", requestObserver);

			var result = messageWithFilesSender.buildNextFileSender();

			assertThat(result).extracting("requestObserver").isSameAs(requestObserver);
		}

		@Test
		void shouldCallAttachmentMetadataMapper() {

			messageWithFilesSender.buildNextFileSender();

			verify(attachmentMetadataMapper).apply(attachment);
		}

		@Test
		void shouldSetAttachmentMetadata() {
			when(attachmentMetadataMapper.apply(any())).thenReturn(attachmentMetadata);

			var result = messageWithFilesSender.buildNextFileSender();

			assertThat(result).extracting("metadata").isSameAs(attachmentMetadata);
		}

	}

	@Nested
	class TestCompleteRequest {

		@Mock
		private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;

		@BeforeEach
		void setup() {
			ReflectionTestUtils.setField(messageWithFilesSender, "requestObserver", requestObserver);
		}

		@Test
		void shouldSetDone() {
			messageWithFilesSender.completeRequest();

			var done = (AtomicBoolean) ReflectionTestUtils.getField(messageWithFilesSender, "done");
			assertThat(done.get()).isTrue();
		}

		@Test
		void shouldCallOnCompleted() {
			messageWithFilesSender.completeRequest();

			verify(requestObserver).onCompleted();
		}
	}

	@Nested
	class TestCancelOnTimeout {

		@Mock
		private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;

		@BeforeEach
		void setup() {
			ReflectionTestUtils.setField(messageWithFilesSender, "requestObserver", requestObserver);
		}

		@Test
		void shouldCallCancelOnRequestObserver() {
			messageWithFilesSender.cancelOnTimeout();

			verify(requestObserver).onError(any(TechnicalException.class));
		}

		@Test
		void shouldCancelResultFuture() {
			messageWithFilesSender.cancelOnTimeout();

			assertThat(messageWithFilesSender.getResultFuture()).isCancelled();
		}
	}

	@Nested
	class TestCancelOnError {

		@Mock
		private CallStreamObserver<GrpcSendBayernIdMessageRequest> requestObserver;

		@BeforeEach
		void setup() {
			ReflectionTestUtils.setField(messageWithFilesSender, "requestObserver", requestObserver);
		}

		@Test
		void shouldCallCancelOnRequestObserver() {
			var exception = new RuntimeException("test");
			messageWithFilesSender.cancelOnError(exception);

			verify(requestObserver).onError(exception);
		}

		@Test
		void shouldCancelResultFuture() {
			messageWithFilesSender.cancelOnError(new TechnicalException("test"));

			assertThat(messageWithFilesSender.getResultFuture()).isCancelled();
		}
	}
}