/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.osi;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.InputStream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.nachrichten.file.AttachmentFile;
import de.ozgcloud.nachrichten.file.AttachmentFileService;
import de.ozgcloud.nachrichten.file.AttachmentFileTestFactory;
import de.ozgcloud.nachrichten.postfach.FileId;

class MessageAttachmentServiceTest {

	private static final FileId FILE_ID = FileId.from(LoremIpsum.getInstance().getWords(1));

	@Spy
	@InjectMocks
	private MessageAttachmentService service;

	@Mock
	private AttachmentFileService attachmentFileService;

	@Nested
	class TestGetMessageAttachment {

		@Mock
		private AttachmentFile attachmentFile;
		@Mock
		private MessageAttachment messageAttachment;

		@BeforeEach
		void init() {
			when(attachmentFileService.getFile(any())).thenReturn(attachmentFile);
			doReturn(MessageAttachmentTestFactory.CONTENT).when(service).getAttachmentContent(any());
			doReturn(messageAttachment).when(service).buildMessageAttachment(any(), any());
		}

		@Test
		void shouldCallGetFile() {
			getMessageAttachment();

			verify(attachmentFileService).getFile(FILE_ID);
		}

		@Test
		void shouldCallGetAttachmentContent() {
			getMessageAttachment();

			verify(service).getAttachmentContent(FILE_ID);
		}

		@Test
		void shouldCallBuildMessageAttachment() {
			getMessageAttachment();

			verify(service).buildMessageAttachment(attachmentFile, MessageAttachmentTestFactory.CONTENT);
		}

		@Test
		void shouldReturnMessageAttachment() {
			var result = getMessageAttachment();

			assertThat(result).isSameAs(messageAttachment);
		}

		private MessageAttachment getMessageAttachment() {
			return service.getMessageAttachment(FILE_ID);
		}
	}

	@Nested
	class TestBuildMessageAttachment {

		@Test
		void shouldBuildMessageAttachment() {
			var attachment = buildMessageAttachment();

			assertThat(attachment).usingRecursiveComparison().isEqualTo(MessageAttachmentTestFactory.create());
		}

		private MessageAttachment buildMessageAttachment() {
			return service.buildMessageAttachment(AttachmentFileTestFactory.create(), MessageAttachmentTestFactory.CONTENT);
		}
	}

	@Nested
	class TestGetAttachmentContent {

		private static final byte[] CONTENT = MessageAttachmentTestFactory.CONTENT.getBytes();

		@BeforeEach
		void init() {
			doReturn(CONTENT).when(service).getContent(any(FileId.class));
			doReturn(MessageAttachmentTestFactory.CONTENT).when(service).encodeAttachmentContent(any());
		}

		@Test
		void shouldCallGetContent() {
			service.getAttachmentContent(FILE_ID);

			verify(service).getContent(FILE_ID);
		}

		@Test
		void shouldCallEncodeAttachmentContent() {
			service.getAttachmentContent(FILE_ID);

			verify(service).encodeAttachmentContent(CONTENT);
		}

		@Test
		void shouldReturnContent() {
			var result = service.getAttachmentContent(FILE_ID);

			assertThat(result).isEqualTo(MessageAttachmentTestFactory.CONTENT);
		}
	}

	@Nested
	class TestGetContent {

		@Test
		void shouldCallGetFileContent() {
			when(attachmentFileService.getFileContent(any())).thenReturn(new ByteArrayInputStream(MessageAttachmentTestFactory.CONTENT.getBytes()));

			service.getContent(FILE_ID);

			verify(attachmentFileService).getFileContent(FILE_ID);
		}

		@Test
		void shouldReturnContent() {
			when(attachmentFileService.getFileContent(any())).thenReturn(new ByteArrayInputStream(MessageAttachmentTestFactory.CONTENT.getBytes()));

			var result = service.getContent(FILE_ID);

			assertThat(result).isEqualTo(MessageAttachmentTestFactory.CONTENT.getBytes());
		}
	}

	@Nested
	class TestEncodeAttachmentContent {

		@Test
		void shouldEncodeContent() {
			var result = service.encodeAttachmentContent(MessageAttachmentTestFactory.DECODED_CONTENT.getBytes());

			assertThat(result).isEqualTo(MessageAttachmentTestFactory.CONTENT);
		}
	}

	@Nested
	class TestPersistAttachment {

		private static final String ATTACHMENT_FILE_ID = LoremIpsum.getInstance().getWords(1);
		private static final MessageAttachment ATTACHMENT = MessageAttachmentTestFactory.create();

		@Mock
		private AttachmentFile attachmentFile;
		@Mock
		private InputStream inputStream;

		@BeforeEach
		void init() {
			doReturn(attachmentFile).when(service).buildAttachmentFile(any(), any());
			when(attachmentFileService.uploadAttachmentFile(any(), any())).thenReturn(ATTACHMENT_FILE_ID);
		}

		@Test
		void shouldCallBuildAttachmentFile() {
			persistAttachment();

			verify(service).buildAttachmentFile(AttachmentFileTestFactory.VORGANG_ID, ATTACHMENT);
		}

		@Test
		void shouldCallCreateAttachmentFile() {
			persistAttachment();

			verify(attachmentFileService).uploadAttachmentFile(attachmentFile, MessageAttachmentTestFactory.CONTENT);
		}

		@Test
		void shouldReturnAttachmentFileId() {
			var result = persistAttachment();

			assertThat(result).isEqualTo(ATTACHMENT_FILE_ID);
		}

		private String persistAttachment() {
			return service.persistAttachment(AttachmentFileTestFactory.VORGANG_ID, ATTACHMENT);
		}
	}

	@Nested
	class TestBuildAttachmentFile {

		@Test
		void shouldMapFileName() {
			var attachmentFile = buildAttachmentFile();

			assertThat(attachmentFile).usingRecursiveComparison().ignoringFields("contentType").isEqualTo(AttachmentFileTestFactory.create());
		}

		private AttachmentFile buildAttachmentFile() {
			return service.buildAttachmentFile(AttachmentFileTestFactory.VORGANG_ID, MessageAttachmentTestFactory.create());
		}
	}
}
