/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.osi;

import static de.ozgcloud.nachrichten.postfach.osi.MessageTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.client.match.MockRestRequestMatchers.*;
import static org.springframework.test.web.client.response.MockRestResponseCreators.*;

import java.util.stream.Collectors;

import org.assertj.core.api.InstanceOfAssertFactories;
import org.assertj.core.util.Arrays;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.http.HttpMethod;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.test.util.ReflectionTestUtils;
import org.springframework.test.web.client.ExpectedCount;
import org.springframework.test.web.client.MockRestServiceServer;
import org.springframework.test.web.client.ResponseActions;
import org.springframework.web.client.RestTemplate;

import de.ozgcloud.nachrichten.postfach.NotConfiguredException;
import de.ozgcloud.nachrichten.postfach.PostfachBadRequestException;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachrichtTestFactory;

class OsiPostfachRemoteServiceTest {

	private static final String TEST_URL = "https://testUrl.local";
	private static final String TEST_API_KEY = "dummyApiKey";
	private static final String TEST_API_REALM = "dummyApiRealm";

	@InjectMocks
	private OsiPostfachRemoteService service;

	@Mock
	private OsiPostfachMessageMapper mapper;

	@Spy
	private RestTemplate restTemplate;

	@Spy
	private OsiPostfachProperties properties = new OsiPostfachProperties();
	{
		properties.setUrl(TEST_URL);
		properties.setRealm(TEST_API_REALM);
		properties.setKey(TEST_API_KEY);
	}

	private MockRestServiceServer mockServer;

	@BeforeEach
	void initMockServer() {
		mockServer = MockRestServiceServer.createServer(restTemplate);
	}

	@Nested
	class TestGetAllMessages {

		@BeforeEach
		void init() {
			prepareMockServer();
		}

		@Test
		void shouldCallRestTemplateExchange() {
			service.getAllMessages();

			mockServer.verify();
		}

		@Test
		void shouldFillMessage() {
			when(mapper.toPostfachNachricht(any())).thenReturn(PostfachNachrichtTestFactory.create());

			var messages = service.getAllMessages();

			assertThat(messages).hasSize(1).first().usingRecursiveComparison().isEqualTo(PostfachNachrichtTestFactory.create());
		}

		@Test
		void shouldThrowExceptionIfNotConfigured() {
			properties.setUrl(null);

			assertThrows(NotConfiguredException.class, () -> service.getAllMessages());
		}

		@Test
		void shouldHaveAttachment() {
			var messages = Arrays.array(MessageTestFactory.create());
			doReturn(ResponseEntity.ok().body(messages)).when(restTemplate).exchange(TEST_URL, HttpMethod.GET, null, Message[].class);
			when(mapper.toPostfachNachricht(messages[0])).thenReturn(PostfachNachrichtTestFactory.create());

			var postfachNachrichten = service.getAllMessages();

			assertThat(postfachNachrichten).hasSize(1).first()
					.extracting(PostfachNachricht::getAttachments, InstanceOfAssertFactories.LIST)
					.hasSize(1)
					.usingRecursiveComparison().isEqualTo(PostfachNachrichtTestFactory.ATTACHMENTS);
		}

		@Test
		void shouldCallMapper() {
			prepareMockServer();

			service.getAllMessages().collect(Collectors.toList());

			verify(mapper).toPostfachNachricht(any());
		}

		private void prepareMockServer() {
			mockServer.expect(ExpectedCount.once(), requestTo(TEST_URL))
					.andExpect(method(HttpMethod.GET))
					.andRespond(withSuccess(MessageJsonReplyTestFactory.create(), MediaType.APPLICATION_JSON));
		}

	}

	@Nested
	class TestSendMessage {

		private final PostfachNachricht message = PostfachNachrichtTestFactory.create();

		@Test
		void shouldCallRestTemplateExchange() {
			prepareMockServer();

			service.sendMessage(message);

			mockServer.verify();
		}

		@Test
		void shouldThrowExceptionUrlIfNotConfigured() {
			prepareMockServer();

			properties.setUrl(null);

			assertThrows(NotConfiguredException.class, () -> service.sendMessage(message));
		}

		@Test
		void shouldThrowExceptionKeyIfNotConfigured() {
			prepareMockServer();

			properties.setKey(null);

			assertThrows(NotConfiguredException.class, () -> service.sendMessage(message));
		}

		@Test
		void shouldThrowExceptionIfNoRestTemplate() {
			ReflectionTestUtils.setField(service, "restTemplate", null);

			assertThrows(NotConfiguredException.class, () -> service.sendMessage(message));
		}

		@Test
		void shouldThrowBadRequestExceptionOnHttpClientError() {
			prepareMockServer().andRespond(withStatus(HttpStatus.BAD_REQUEST));

			assertThrows(PostfachBadRequestException.class, () -> service.sendMessage(message));
		}

		@Test
		void shouldCallMapper() {
			prepareMockServer();

			service.sendMessage(message);

			verify(mapper).toOsiMessage(any());
		}

		@Nested
		class TestResponse {

			@Test
			void shouldThrowServerProcessExceptionOnResponseBodyFalse() {
				prepareMockServerWithResponseBody(false);

				assertThrows(OsiPostfachServerProcessException.class, () -> service.sendMessage(message));

				mockServer.verify();
			}

			void prepareMockServerWithResponseBody(boolean body) {
				mockServer.expect(ExpectedCount.once(), requestTo(TEST_URL))
						.andExpect(method(HttpMethod.POST))
						.andRespond(withSuccess(String.valueOf(body), MediaType.APPLICATION_JSON));
			}
		}

		@DisplayName("rest request body")
		@Nested
		class TestRestRequestBody {

			@Test
			void shouldHaveVorgangIdAsSequenceNumber() {
				var postfachNachricht = PostfachNachrichtTestFactory.create();
				when(mapper.toOsiMessage(postfachNachricht)).thenReturn(MessageTestFactory.create());

				prepareMockServer().andExpect(jsonPath("$.sequenceNumber").value(VORGANG_ID));

				service.sendMessage(postfachNachricht);
			}

			@Test
			void shouldHavePostfachIdAsNameIdentifier() {
				var postfachNachricht = PostfachNachrichtTestFactory.create();
				when(mapper.toOsiMessage(postfachNachricht)).thenReturn(MessageTestFactory.create());

				prepareMockServer().andExpect(jsonPath("$.nameIdentifier").value(POSTFACH_ID));

				service.sendMessage(postfachNachricht);
			}

			@Test
			void shouldHaveSubject() {
				var postfachNachricht = PostfachNachrichtTestFactory.create();
				when(mapper.toOsiMessage(postfachNachricht)).thenReturn(MessageTestFactory.create());

				prepareMockServer().andExpect(jsonPath("$.subject").value(SUBJECT));

				service.sendMessage(postfachNachricht);
			}

			@Test
			void shouldHaveMailBodyAsBody() {
				var postfachNachricht = PostfachNachrichtTestFactory.create();
				when(mapper.toOsiMessage(postfachNachricht)).thenReturn(MessageTestFactory.create());

				prepareMockServer().andExpect(jsonPath("$.body").value(MAIL_BODY));

				service.sendMessage(postfachNachricht);
			}

			@Test
			void shouldHaveNumericReplyOptionAsReplyAction() {
				var postfachNachricht = PostfachNachrichtTestFactory.create();
				when(mapper.toOsiMessage(postfachNachricht)).thenReturn(MessageTestFactory.create());

				prepareMockServer().andExpect(jsonPath("$.replyAction").value(REPLY_OPTION.toValue()));

				service.sendMessage(postfachNachricht);
			}
		}

		private ResponseActions prepareMockServer() {
			var responseAction = mockServer.expect(ExpectedCount.once(), requestTo(TEST_URL))
					.andExpect(method(HttpMethod.POST));

			responseAction.andRespond(withSuccess());
			return responseAction;
		}
	}

	@Nested
	class TestDeleteMessage {

		private final String DELETE_URL = TEST_URL + "?messageGuid=" + MessageTestFactory.MESSAGE_ID;

		@Test
		void shouldCallRestTemplateExchange() {
			prepareMockServer();

			service.deleteMessage(MessageTestFactory.MESSAGE_ID);

			mockServer.verify();
		}

		@Test
		void shouldThrowExceptionIfNotConfigured() {
			prepareMockServer();

			properties.setUrl(null);

			assertThrows(NotConfiguredException.class, () -> service.deleteMessage(MessageTestFactory.MESSAGE_ID));
		}

		private void prepareMockServer() {
			mockServer.expect(ExpectedCount.once(), requestTo(DELETE_URL))
					.andExpect(method(HttpMethod.DELETE))
					.andRespond(withSuccess());
		}
	}

	@Nested
	class TestGetPostfach {

		@Test
		void shouldHasTyp() {
			var postfachType = service.getPostfachType();

			assertThat(postfachType).isEqualTo(OsiPostfachRemoteService.POSTFACH_TYPE);
		}

		@Test
		void shouldAllowReplys(){
			var isReplyAllowed = service.isReplyAllowed();

			assertThat(isReplyAllowed).isTrue();
		}
	}
}